#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module contains classes used in the regression

import stat_functions as stat
import numpy as np
import time
import threading
import debug
import functions as fu
import calculus_functions as cf
import model_parser
import calculus
import copy
import loglikelihood as logl
from scipy import sparse as sp
import arguments


min_AC=0.000001

#todo: split up panel in sub classes
def posdef(a,da):
	return list(range(a,a+da)),a+da

class panel:
	def __init__(self,datainput,settings,pqdkm=None):

		self.input=datainput
		self.options=settings
		self.dataframe=datainput.dataframe
		if not pqdkm is None:
			self.pqdkm=pqdkm
		else:
			self.pqdkm=settings.pqdkm.value
		self.initial_defs()
		self.arrayize()
		self.masking()
		self.lag_variables()
		self.final_defs()
		

	def initial_defs(self):
		if np.all(np.var(self.input.Y,0)==0):
			raise RuntimeError("No variation in Y")
		p,q,d,k,m=self.pqdkm
		self.max_lags=self.input.max_lags
		self.lost_obs=max((p,q,self.max_lags))+max((m,k,self.max_lags))+d#+3
		self.nW,self.nZ,self.n_beta=len(self.input.W.columns),len(self.input.Z.columns),len(self.input.X.columns)
		self.define_h_func()
		self.Ld_inv=None
		if self.input.IDs is None:
			self.options.fixed_random_group_eff.value=0
		if self.input.IDs is None:
			self.options.fixed_random_group_eff.value=0		
		self.m_zero = False
		if  m==0 and k>0:
			self.m_zero = True
			p,q,d,k,m=self.pqdkm
			self.pqdkm=p,q,d,k,1
			
	def masking(self):
		
		#"initial observations" mask: 
		a=np.array([self.date_counter<self.T_arr[i] for i in range(self.N)])# sets observations that shall be zero to zero by multiplying it with the arrayized variable
		self.a=[None,None]
		self.a.extend([a.reshape(list(a.shape)[:-1]+[1]*i) for i in range(5)])
		#"after lost observations" masks: 
		self.T_i=np.sum(self.included[3],1).reshape((self.N,1,1))#number of observations for each i
		self.T_i=self.T_i+(self.T_i<=0)#ensures minimum of 1 observation in order to avoid division error. If there are no observations, averages will be zero in any case	
		self.N_t=np.sum(self.included[3],0).reshape((1,self.max_T,1))#number of observations for each t
		self.N_t=self.N_t+(self.N_t<=0)#ensures minimum of 1 observation in order to avoid division error. If there are no observations, averages will be zero in any case	
		self.group_var_wght=1-1/np.maximum(self.T_i-1,1)
		
	def final_defs(self):
		self.W_a=self.W*self.a[3]
		self.tot_lost_obs=self.lost_obs*self.N
		self.NT=np.sum(self.included[3])
		self.NT_before_loss=self.NT+self.tot_lost_obs
		if not hasattr(self,'n_dates'):
			self.number_of_RE_coef=0
			self.number_of_RE_coef_in_variance=0
			self.options.fixed_random_group_eff.set(0)
			self.options.fixed_random_time_eff.set(0)
			self.options.fixed_random_variance_eff.set(0)
		else:
			self.number_of_RE_coef=self.N*(self.options.fixed_random_group_eff.value>0)+self.n_dates*(self.options.fixed_random_time_eff.value>0)
			self.number_of_RE_coef_in_variance=(self.N*(self.options.fixed_random_group_eff.value>0)
												+self.n_dates*(self.options.fixed_random_time_eff.value>0))*(self.options.fixed_random_variance_eff.value>0)
		self.args=arguments.arguments(self)
		self.df=self.NT-self.args.n_args-self.number_of_RE_coef-self.number_of_RE_coef_in_variance
		self.set_instrumentals()
		self.tobit()

	def lag_variables(self):
		T=self.max_T
		d=self.pqdkm[2]
		self.I=np.diag(np.ones(T))
		#differencing operator:
		if d==0:
			return
		L0=np.diag(np.ones(T-1),-1)
		Ld=(self.I-L0)
		Ld_inv=np.tril(np.ones((T,T)))
		for i in range(1,d):
			Ld=cf.dot(self.I-L0,Ld)
			Ld_inv=np.cumsum(Ld_inv,0)
		self.Ld_inv=Ld_inv
		#multiplication:
		self.X=self.lag_variable(self.X, Ld, d, True)
		self.Y=self.lag_variable(self.Y, Ld, d, False)
		if not self.Z is None:
			self.Z=self.lag_variable(self.Z, Ld, d, True)
				
	def lag_variable(self,X,Ld,d,recreate_intercept):
		X_out=cf.dot(Ld,X)*self.a[3]
		if self.input.has_intercept and recreate_intercept:
			X_out[:,:,0]=1
		X_out[:,:d]=0
		return X_out

	def params_ok(self,args):
		a=self.q_sel,self.p_sel,self.M_sel,self.K_sel
		for i in a:
			if len(i)>0:
				if np.any(np.abs(args[i])>0.999):
					return False
		return True
	
	def set_instrumentals(self):
		if self.input.Z.shape[1]==1:
			self.XIV=self.X
		else:
			self.XIV=self.X=self.Z
		return
		ll=logl.LL(self.args.args_init,self)
		ll.standardize()
		Z_st,Z_st_long=ll.standardize_variable(self.Z)
		ZZ=cf.dot(Z_st,Z_st)
		ZZInv=np.linalg.inv(ZZ)
		ZX=cf.dot(Z_st,ll.X_st)
		ZZInv_ZX=cf.dot(ZZInv, ZX)
		self.XIV=cf.dot(self.Z, ZZInv_ZX)#using non-normalized first, since XIV should be unnormalized.	

	def subtract_means(self,X,Y,Z):
		subtract=self.options.subtract_means.value
		if not subtract:
			return X,Y,Z
		X=X-np.mean(X,0)
		if self.input.has_intercept:
			X[:,0]=1
		Y=Y-np.mean(Y,0)
		if self.input.Z.shape[1]==1:
			return X,Y,Z
		Z=Z-np.mean(Z,0)
		Z[:,0]=1		
		
	def is_single(self):
		IDs,t=self.input.IDs_num,self.input.timevar_num
		try:
			if (np.all(IDs.iloc[0]==IDs) or np.all(t.iloc[0]==t)):
				return True
		except:
			return True
		return False
		
		
	def arrayize(self):
		"""Splits X and Y into an arry of equally sized matrixes rows equal to the largest for each IDs"""

		X, Y, W, IDs,timevar,Z=[to_numpy(i) for i in 
								(self.input.X, self.input.Y, self.input.W, self.input.IDs_num,self.input.timevar_num,self.input.Z)]
		X,Y,Z=self.subtract_means(X,Y,Z)
		NT,k=X.shape
		self.total_obs=NT
		if self.is_single():
			if not np.all(timevar[:,0]==np.sort(timevar[:,0])):#remove in production
				raise RuntimeError("The time variable is not sorted!?!")			
			self.X=np.array(X.reshape((1,NT,k)))
			self.Y=np.array(Y.reshape((1,NT,1)))
			NT,kW=W.shape
			self.W=np.array(W.reshape((1,NT,kW)))
			NT,kZ=Z.shape
			self.Z=None
			if Z.shape[1]>1:#instrumental variables used
				self.Z=np.array(Z.reshape((1,NT,kZ)))
			self.time_map=None
			self.map=np.arange(NT).reshape(1,NT)
			self.N=1
			self.max_T=NT
			self.T_arr=np.array([[NT]])
			self.date_counter=np.arange(self.max_T).reshape((self.max_T,1))
			included=np.array([(self.date_counter>=self.lost_obs)*(self.date_counter<self.T_arr[i]) for i in range(self.N)])
			self.date_count_mtrx=None
			self.date_count=None
			self.idincl=np.array([True])
			if not np.all(timevar[:,0]==np.sort(timevar[:,0])):#remove in production
				raise RuntimeError("The arrayize procedure has unsorted the time variable!?!")					
		else:
			sel,ix=np.unique(IDs,return_index=True)
			N=len(sel)
			sel=(IDs.T==sel.reshape((N,1)))
			T=np.sum(sel,1)
			self.max_T=np.max(T)
			self.idincl=T>self.lost_obs+self.options.min_group_df.value
			self.X=arrayize(X, N,self.max_T,T, self.idincl,sel)
			self.Y=arrayize(Y, N,self.max_T,T, self.idincl,sel)
			self.W=arrayize(W, N,self.max_T,T, self.idincl,sel)
			self.Z=arrayize(Z, N,self.max_T,T, self.idincl,sel)		
			self.N=np.sum(self.idincl)
			varmap=np.arange(NT).reshape(NT,1)
			self.map=arrayize(varmap, N,self.max_T,T, self.idincl,sel,dtype=int).reshape((self.N,self.max_T))
			self.T_arr=T[self.idincl].reshape((self.N,1))
			self.date_counter=np.arange(self.max_T).reshape((self.max_T,1))
			included=np.array([(self.date_counter>=self.lost_obs)*(self.date_counter<self.T_arr[i]) for i in range(self.N)])
			if len(included)<5:
				raise RuntimeError(f"{len(included)} valid observations, cannot perform panel analysis. Try without panel (don't specify ID and time)")
			self.get_time_map(timevar, self.N,T, self.idincl,sel,included)
			
			if np.sum(self.idincl)<len(self.idincl):
				idname=self.input.IDs.columns[0]
				idremoved=np.array(self.input.IDs)[ix,0][self.idincl==False]
				s=fu.formatarray(idremoved,90,', ')
				print(f"The following {idname}s were removed because of insufficient observations:\n %s" %(s))
			#remove in production. Checking sorting:
			tvar=arrayize(timevar, N,self.max_T,T, self.idincl,sel)
			a=[tvar[i][0][0] for i in self.date_map]
			if not np.all(a==np.sort(a)):	
				raise RuntimeError("It seems the data is not properly sorted on time")
		
		zeros=np.zeros((self.N,self.max_T,1))
		ones=np.ones((self.N,self.max_T,1))			
		self.included=[None,None]
		self.zeros=[None,None]
		self.ones=[None,None]
		self.included.extend([included.reshape(list(included.shape)[:-1]+[1]*i) for i in range(5)])		
		self.zeros.extend([zeros.reshape(list(zeros.shape)[:-1]+[1]*i) for i in range(5)])	
		self.ones.extend([ones.reshape(list(ones.shape)[:-1]+[1]*i) for i in range(5)])	
		

	def tobit(self):
		"""Sets the tobit threshold"""
		tobit_limits=self.options.tobit_limits.value
		if tobit_limits is None:
			return
		if len(tobit_limits)!=2:
			print("Warning: The tobit_limits argument must have lenght 2, and be on the form [floor, ceiling]. None is used to indicate no active limit")
		if (not (tobit_limits[0] is None)) and (not( tobit_limits[1] is None)):
			if tobit_limits[0]>tobit_limits[1]:
				raise RuntimeError("floor>ceiling. The tobit_limits argument must have lenght 2, and be on the form [floor, ceiling]. None is used to indicate no active limit")
		g=[1,-1]
		self.tobit_I=[None,None]
		self.tobit_active=[False, False]#lower/upper threshold
		desc=['tobit_low','tobit_high']
		for i in [0,1]:
			self.tobit_active[i]=not (tobit_limits[i] is None)
			if self.tobit_active[i]:
				if np.sum((g[i]*self.Y<g[i]*tobit_limits[i])*self.a[3]):
					print("Warning: there are observations of self.Y outside the non-censored interval. These will be ignored.")
				I=(g[i]*self.Y<=g[i]*tobit_limits[i])*self.a[3]
				self.Y[I]=tobit_limits[i]
				if np.var(self.Y)==0:
					raise RuntimeError("Your tobit limits are too restrictive. All observationss would have been cencored. Cannot run regression with these limits.")
				self.tobit_I[i]=I
				if np.sum(I)>0 and np.sum(I)<self.NT and False:#avoiding singularity #Not sure why this is here, shuld be deleted?
					self.X=np.concatenate((self.X,I),2)
					self.input.X_names.append(desc[i])
	
	def get_time_map(self,timevar, N,T_count, idincl,sel,incl):
		if timevar is None:
			return None
		N,T,k=incl.shape
		unq,ix=np.unique(timevar,return_inverse=True)
		t=arrayize(np.array(ix).reshape((len(timevar),1)), 
				   N,self.max_T,T_count, idincl,sel,int)#maps N,T -> unique date
		grp_cnt=incl*np.arange(N).reshape(N,1,1)
		t_cnt=incl*np.arange(T).reshape(1,T,1)
		incl=incl[:,:,0]
		t=np.concatenate((t[incl],  grp_cnt[incl], t_cnt[incl]),1)
		a=np.argsort(t[:,0])
		t=t[a]#three columns: unique date index, group number, day sequence
		
	
		tid=t[:,0]#unique date
		t_map=[[] for i in range(np.max(tid)+1)]#all possible unique dates
		for i in range(len(tid)):
			t_map[tid[i]].append(t[i,1:])#appends group and day sequence
		t_map_tuple=[]
		tcnt=[]
		self.date_count_mtrx=np.zeros((N,T,1))
		for i in range(len(t_map)):
			a=np.array(t_map[i]).T#group and day sequence for unique date i
			if len(a):
				m=(tuple(a[0]),tuple(a[1]))#group and day sequence reference tuple
				n_t=len(a[0])#number of groups at this unique date
				t_map_tuple.append(m)	#group and day reference tuple for the data matrix, for each unique date
				tcnt.append(n_t) #count of groups at each date
				self.date_count_mtrx[m]=n_t#timeseries matrix of the group count
				
		
		#A full random effects calculation is infeasible because of complexity and computing costs. 
		#A quazi random effects weighting is used. It  is more conservative than the full
		#RE weight theta=1-sd_pooled/(sd_pooled+sd_within/T)**0.5
		#If the weights are too generous, the RE adjustment may add in stead of reducing noise. 
		n=len(tcnt)
		self.n_dates=n
		self.date_count=np.array(tcnt).reshape(n,1,1)
		self.date_map=t_map_tuple
		
		
	def get_time_map2(self,timevar, N,T_count, idincl,sel):
		if timevar is None:
			return None
		unq,ix=np.unique(timevar,return_inverse=True)
		n_dates=len(unq)
		t=arrayize(np.array(ix).reshape((n_dates,1)), N,self.max_T,T_count, idincl,sel,int)#maps N,T -> unique date
		N,T,k=t.shape

		
		t_map=[[] for i in range(n_dates)]#all possible unique dates
		for i in range(len(tid)):
			t_map[tid[i]].append(t[i,1:])#appends group and day sequence
		t_map_tuple=[]
		tcnt=[]
		self.date_count_mtrx=np.zeros((N,T,1))
		for i in range(len(t_map)):
			a=np.array(t_map[i]).T#group and day sequence for unique date i
			if len(a):
				m=(tuple(a[0]),tuple(a[1]))#group and day sequence reference tuple
				n_t=len(a[0])#number of groups at this unique date
				t_map_tuple.append(m)	#group and day sequence reference tuple, for each unique date
				tcnt.append(n_t) #count of groups at each date
				self.date_count_mtrx[m]=n_t#timeseries matrix of the group count
				
		
		#A full random effects calculation is infeasible because of complexity and computing costs. 
		#A quazi random effects weighting is used. It  is more conservative than the full
		#RE weight theta=1-sd_pooled/(sd_pooled+sd_within/T)**0.5
		#If the weights are too generous, the RE adjustment may add in stead of reducing noise. 
		n=len(tcnt)
		self.n_dates=n
		self.date_count=np.array(tcnt).reshape(n,1,1)
		self.date_map=t_map_tuple
	
	
	def define_h_func(self):
		h_def="""
def h(e,z):
	e2			=	e**2+1e-5
	h_val		=	np.log(e2)	
	h_e_val		=	2*e/e2
	h_2e_val	=	2/e2-4*e**2/e2**2

	return h_val,h_e_val,h_2e_val,None,None,None
		"""	
		h_definition=self.options.h_function.value
		if h_definition is None:
			h_definition=h_def
		d=dict()
		try:
			exec(h_definition,globals(),d)
			ret=d['h'](1,1)
			if len(ret)!=6:
				raise RuntimeError("""Your custom h-function must return exactly six arguments
				(h, dh/dx and ddh/dxdx for both e and z. the z return values can be set to None)""")
			self.h_def=h_definition
		except Exception as e:
			print('Something is wrong with your custom function, default is used:'+ str(e))
			exec(h_def,globals(),d)
			self.h_def=h_def
		
		self.z_active=True
		for i in ret[3:]:
			self.z_active=self.z_active and not (i is None)	
		
		
	def mean(self,X,axis=None):
		dims=list(X.shape)
		dims[2:]=[1]*(len(dims)-2)
		#X=X*self.included.reshape(dims)
		if axis is None:
			return np.sum(X)/self.NT
		if axis==1:
			dims.pop(1)
			return np.sum(X,1)/self.T_i.reshape(dims)
		if axis==0:
			dims.pop(0)
			return np.sum(X,0)/self.N_t.reshape(dims)
		if axis==(0,1):
			return np.sum(np.sum(X,0),0)/self.NT
			
	def var(self,X,axis=None,k=1,mean=None,included=None):
		dims=list(X.shape)
		dims_m=np.array(X.shape)
		dims[2:]=[1]*(len(dims)-2)	
		if included is None:
			a=self.included[len(dims)]
		else:
			a=included
		if mean is None:
			m=self.mean(X*a, axis)
		else:
			m=mean

		if axis==None:
			Xm=(X-m)*a
			return np.sum(Xm**2)/(self.NT-k)

		if axis==1:
			dims_m[1]=1
			m=m.reshape(dims_m)
			Xm=(X-m)*a
			dims.pop(1)
			return np.sum((Xm)**2,1)/np.maximum(self.T_i-k,1).reshape(dims)
		if axis==0:
			dims_m[0]=1		
			m=m.reshape(dims_m)
			Xm=(X-m)*a
			dims.pop(0)
			return np.sum((Xm)**2,0)/np.maximum(self.N_t-k,1).reshape(dims)
		if axis==(0,1):
			dims_m[0:2]=1
			m=m.reshape(dims_m)
			Xm=(X-m)*a			
			return np.sum((Xm)**2,axis)/(self.NT-k)
	
def arrayize(X,N,max_T,T,idincl,sel,dtype=None):
	if X is None:
		return None
	NT,k=X.shape
	if dtype is None:
		Xarr=np.zeros((N,max_T,k))
	else:
		Xarr=np.zeros((N,max_T,k),dtype=dtype)
	T_used=[]
	k=0
	for i in range(len(sel)):
		if idincl[i]:
			Xarr[k,:T[i]]=X[sel[i]]
			k+=1
	Xarr=Xarr[:k]
	return Xarr






def to_numpy(x):
	x=np.array(x)
	if len(x.shape)==2:
		return x
	return x.reshape((len(x),1))



import tkinter as tk
import time
import threading


def main():
    top = tk.Tk()
    tex = tk.Text(master=top)
    tex.pack(side=tk.TOP)
    bop = tk.Frame()
    bop.pack(side=tk.BOTTOM,fill=tk.X)
    b = tk.Button(bop, text='Visual', command=cbc("hey adfa", tex))
    b.pack(side=tk.LEFT,fill=tk.X)
    b = tk.Button(bop, text='Columns', command=cbc("ho adfa", tex))
    b.pack(side=tk.LEFT)    
    
    tk.Button(bop, text='Exit', command=top.destroy).pack(side=tk.RIGHT)
    threadit(tex,dosomething,(tex,))
    top.mainloop()  

def cbc(string, tex):
    return lambda : callback(string, tex)


def callback(string, tex):
    tex.insert(tk.END, string)
    tex.see(tk.END)             # Scroll if necessary
    
    
def threadit(tex,func,args):
    t = threading.Thread(target=func,args=args)
    t.start()
    
def dosomething(tex):
    for i in range(20):
        time.sleep(5)
        tex.insert(tk.END, "id: %s" %(i,))



main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

DEFAULT_INTERCEPT_NAME='Intercept'
VAR_INTERCEPT_NAME='log_variance_constant'
INSTRUMENT_INTERCEPT_NAME='instrument_intercept'
CONST_NAME='one'
NUMERIC_TAG="_numeric"

import numpy as np
import functions as fu
import loaddata
import numpy.lib.recfunctions as rfn
import pandas as pd

def get_variables(ip,df,model_string,IDs,timevar,heteroscedasticity_factors,instruments,settings,pool=(None,'mean')):
	print ("Analyzing variables ...")
	if not type(df)==pd.DataFrame:
		raise RuntimeError('The dataframe supplied is not a pandas dataframe. Only pandas dataframes are supported.')
	if CONST_NAME in df:
		print(f"Warning: The name {CONST_NAME} is reserved for the constant 1."
			  f"The variable with this name will be overwritten and set to 1")
	df=pool_func(df,pool)
	
	df[CONST_NAME]=1
	df[DEFAULT_INTERCEPT_NAME]     = df[CONST_NAME]
	df[VAR_INTERCEPT_NAME]         = df[CONST_NAME]
	df[INSTRUMENT_INTERCEPT_NAME]  = df[CONST_NAME]
	
	IDs     =get_names(IDs, 'IDs')
	timevar =get_names(timevar, 'timevar')
	IDs_num,     IDs      = handle_IDs(ip,df,IDs)
	timevar_num, timevar  = handle_time(ip,df,timevar)	
	sort=[i for i in [IDs,timevar] if not i==[]]
	if len(sort):
		df=df.sort_values([IDs_num[0],timevar[0]])
	pd_panel=df.groupby(IDs_num)
	
	W=get_names(heteroscedasticity_factors,'heteroscedasticity_factors',True, VAR_INTERCEPT_NAME)
	Z=get_names(instruments,'instruments',True,INSTRUMENT_INTERCEPT_NAME)
	
	try:
		Y,X=parse_model(model_string, settings)
	except:
		RuntimeError("The model_string must be on the form Y~X1+X2+X3")	
	if X==[] or Y==['']:
		raise RuntimeError("No independent or dependent variables specified")		
	
	x=IDs_num+IDs+timevar+timevar_num+W+Z+Y+X
	df,ip.max_lags=eval_variables(x, df, pd_panel)
	df=df[x]
	n=len(df)
	df=df.dropna()
	ip.lost_na_obs=(n-len(df))-ip.max_lags
	const={}
	for x,add_intercept,num in [
				('IDs_num',False,True),('timevar_num',False,True),
				('IDs',False,False),('timevar',False,False),
			  	('W',True,True),('Z',True,True),('Y',False,True),
			  	('X',settings.add_intercept.value,True)]:
		ip.__dict__[x], const[x]= check_var(df,locals()[x],x,add_intercept,num)
		ip.__dict__[x+"_names"]=list(ip.__dict__[x].columns)
	ip.dataframe=df
	ip.has_intercept=const['X']

	
	
def pool_func(df,pool):
	x,operation=pool
	if x is None:
		return df
	x=get_names(x, 'pool')
	df=df.groupy(x).agg(operation)
	return df
	
def check_var(df,x,inputtype,add_intercept,numeric):
	if len(x)==0:
		return None,None
	dfx=df[x]
	if not numeric:
		return dfx,None
	const_found=False
	for i in x:
		if ' ' in i:
			raise RuntimeError(f'Spaces are not allowed in variables, but found in the variable {i} from {inputtype}')
		try:
			v=np.var(dfx[i])
		except TypeError as e:
			raise TypeError(f"All variables except time and id must be numeric. {e}")
		if v==0 and const_found:
			if dfx[i].iloc[0]==0:
				print(f"Warning: All values in {i} from {inputtype} are zero, variable dropped")
			else:
				print(f"Warning: {i} from {inputtype} is constant. Variable dropped.")
			dfx=dfx.drop(i,1)
		elif v==0 and not const_found:
			if inputtype=='Y':
				raise RuntimeError('The dependent variable is constant')	
			const_found=True
	return dfx,const_found
	
def eval_variables(x,df,pd_panel):
	lag_obj=lag_object(pd_panel)
	d={'D':lag_obj.diff,'L':lag_obj.lag,'np':np}	
	for i in df.keys():
		d[i]=df[i]
	for i in x:
		df[i]=eval(i,d)
	return df,lag_obj.max_lags

class lag_object:
	def __init__(self,panel):
		self.panel=panel
		self.max_lags=0
		
	def lag(self,variable,lags=1):
		x=self.panel[variable.name].shift(lags)
		self.max_lags=max((self.max_lags,lags))
		return x
	
	def diff(self,variable,lags=1):
		x=self.panel[variable.name].diff(lags)
		self.max_lags=max((self.max_lags,lags))
		return x

def parse_model(model_string,settings):
	for i in ['~','=']:
		if i in model_string:
			split=i
			break
	Y,X=model_string.split(split)
	X=[i.strip() for i in X.split('+')]
	if X==['']:
		X=[]
	if settings.add_intercept.value:
		X=[DEFAULT_INTERCEPT_NAME]+X
	return [Y],X
		
def get_names(x,inputtype,add_intercept=False,intercept_name=None):
	if x is None:
		r=[]
	elif type(x)==str:
		r=[x]
	elif type(x)==pd.DataFrame:		
		r=x.columns
	elif type(x)==pd.Series:
		r=[x.name]
	else:
		raise RuntimeError(f"Input for {inputtype} needs to be a list or tuple of strings, a pandas DataFrame object or a pandas Series object")
	if add_intercept:
		r=[intercept_name]+r
	return r
	
def handle_time(ip,df,x):
	if x==[]:
		return [],[]
	x=x[0]
	try:
		x_dt=pd.to_datetime(df[x])
	except ValueError as e:
		try:
			x_dt=pd.to_numeric(x_dt)
		except ValueError as e:
			raise ValueError(f'Expected date or numeric for {inputtype}, but {x} is not recognized as a date or numeric variable by pandas.')
	x_dt=pd.to_numeric(x_dt)/(24*60*60*1000000000)
	x_int=x_dt.astype(int)
	if len(pd.unique(df[x]))==len(pd.unique(x_int)):
		x_dt=x_int
	df[x+NUMERIC_TAG]=x_dt
	return [x+NUMERIC_TAG],[x]


def handle_IDs(ip,df,x):
	if x==[]:
		return [],[]
	x=x[0]
	ids, ip.IDs_unique = pd.factorize(df[x],True)
	df[x+NUMERIC_TAG]=ids
	#both these are true before next assignment:
	#np.all(ip.IDs_unique[ids]==df[x])
	#np.all(np.arange(len(ids))[ids]==ids)
	return [x+NUMERIC_TAG],[x]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from gui import gui_charts
from gui import gui_scrolltext
from gui import gui_scatter_charts
from tkinter import ttk
from tkinter import font
import functions as fu
import stat_functions as st
import traceback
import output
import os
import numpy as np
from PIL import ImageGrab
import communication

#color map:
bg_normal='white'
bg_selected='dark blue'
fg_normal='black'
fg_disabled='grey'
bg_sub_menu="#fafafa"
			
class output_tab(tk.Frame):
	def __init__(self,window,exe_tab,name,tabs,main_tabs,output_data):
		top_size=22
		self.txt_line_zoom=0
		self.window=window
		name=tabs.gen_name(name)
		self.tabs=tabs
		self.name=name
		tk.Frame.__init__(self,main_tabs)
		
		self.widget=tk.Frame(self)

		self.widget.rowconfigure(0,weight=1)
		self.widget.columnconfigure(0,weight=1)			
		
		
		self.charts = gui_charts.process_charts(window,self.widget,main_tabs,tabs)
		self.progress_bar=bar(self,exe_tab)
		
		self.progress_bar.grid(row=2, sticky=tk.EW)
		
		#adding the output tab to tha main tabs notebook
		top_color='#fcf3d9'
		self.tab=tabs.add(self,name=name,top_text='Regression output',top_color=top_color,skip_buttons=True)
		self.tab.exe_tab=exe_tab
		
		self.scroll_text_frame=tk.Frame(self.widget)
		self.scroll_text_frame.rowconfigure(0)
		self.scroll_text_frame.rowconfigure(1,weight=1)
		self.scroll_text_frame.columnconfigure(0,weight=1)			
		self.selection_frame=selection_frame(self.scroll_text_frame)
		self.tab.widget= gui_scrolltext.ScrollText(self.scroll_text_frame,format_text=False,name='regression',window=window,)
		self.output_text= gui_scrolltext.ScrollText(self.scroll_text_frame,format_text=False,name='regression',window=window,)
		main_tabs.select(self)
		main_tabs.insert('end',tabs.add_tab)
		self.scatter=gui_scatter_charts.scatter_charts(self.scroll_text_frame,tabs)
		self.scatter_norm=gui_scatter_charts.scatter_charts(self.scroll_text_frame,tabs)

		self.statistics=None
		self.reg_table=None
		self.widget.stored_output_data=stored_output()
		if not output_data is None:
			if hasattr(output_data, 'chart_images'):
				self.widget.stored_output_data=output_data
				try:
					self.charts.charts_from_stored(output_data.chart_images)
				except Exception as e:
					print(e)
					print('Chart images not found in temp folder. Change settings to\n'
						  'clean up the temp folder less frequently, in order ot avoid\n'
						  'this problem again\n')
				
		if not 'menu_selections' in self.window.data:
			self.window.data['menu_selections']=dict()
		self.add_buttons(top_size,top_color)
		self.add_formats()
		
		
	def add_formats(self):
		t=self.tab.widget.text_box
		t.configure(font=('Garamond', 12))
		t.tag_configure('bold_G', font=('Garamond', 12,'bold'))
		t.tag_configure('bold_G_underline', font=('Garamond', 12,'bold','underline'))
		t.tag_configure('10p', font=('Garamond', 10))		
		t.tag_configure('space_line', font=('Garamond', 7))
		
	def add_lines(self,zoom):		
		dirname=os.path.dirname(__file__)	
		if zoom==self.txt_line_zoom:
			return
		self.txt_line = tk.PhotoImage(file =  os.path.join(dirname,'img','line.png'))
		self.txt_line=self.txt_line.zoom(zoom,1)
		self.txt_double_line = tk.PhotoImage(file =  os.path.join(dirname,'img','double_line.png'))
		self.txt_double_line=self.txt_double_line.zoom(zoom,1)	
		self.txt_line_zoom=zoom
		
	
	def add_buttons(self,top_size,top_color):
		
		self.button_img=dict()
		
		self.button_img['save']= tk.PhotoImage(file =  fu.join(os.path.dirname(__file__),['img','save.png']),master=self.tab.button_frame)
		self.button_save=tk.Button(self.tab.button_frame_L, image = self.button_img['save'],command=self.save, 
								   highlightthickness=0,bd=0,height=top_size, anchor=tk.E,background=top_color)	
		self.button_save.grid(row=0,column=1)
		
		#selection frame buttons:
		self.font = font.Font(family='Helvetica', size=11, weight=font.BOLD)	
		self.font_small = font.Font(family='Helvetica', size=9, weight=font.BOLD)	

		
		self.menu_buttons={}
		i=0
		
		btnlist=[
			# Main button caption	Sub captions										command								click		sub 				Default
			#																												type main	groups				enabled sub							
			['REGRESSION',			['DIR','CNSTRNTS',['( )', '[ ]', 'disabled:( )'],
										['FLAT', 'STACKED'],
										['NORMAL', 'HTML', 'LATEX', 'RTF','INTERNAL'],
										['JOINED', 'JOINED LONG', 'disabled:JOINED']],	self.print_regression,				'group',	[[2,3,4]],			[0,1,2,3,4]],
			['SCATTER PLOTS',		['NORMALIZED'],										self.show_scatter,					'group', 	None,				[1]],
			[' CORREL ',			['NORMALIZED',[' CORREL ',' COVAR '],
										['VARIABLES','ESTIMATES','RAW MOMENTS']],		self.print_correl,					'group', 	[[1],[2]],			[1,2]],
			['DIAGNOSTICS',			['SAMP.SIZE','TESTS'],								self.print_stats,					'group',	[None],				[None]],
			['DESCRIPTIVES',		['EXP','LN','ORIGINAL'],							self.print_descriptive_statistics,	'group', 	[[0,1,2]],			[2]],
			['DIGITS',				list(range(10))+[16]+['SCI'],						self.print,							None,		[range(12)],		[5]],
			['DISTRIBUTION CHARTS',	None,												self.show_dist_charts,				'toggle',	None,				[None]]
		]
		#"disabled:" must come last in the sublist
		for caption, captions_sub,cmd,click_type,group_type_sub,def_sub in btnlist:
			self.menu_buttons[caption]=menu_button(self, i, caption, cmd,captions_sub,click_type,group_type_sub,def_sub)	
			i+=1

		self.no_print_menus=['DIGITS','FORMAT']
		self.widget.grid(row=1,column=0,sticky=tk.NSEW)
		self.tab.widget.grid(column=0,row=1,sticky=tk.NSEW)
		self.scroll_text_frame.grid(column=0,row=0,sticky=tk.NSEW)
		self.selection_frame.grid(0,0)
		self.charts.grid(column=1,row=0,sticky=tk.NS)
		self.get_stored()
		self.menu_buttons['REGRESSION'].click()
	
	
	def get_stored(self):
		"Obtains stored states of the buttons"
		if len(self.window.data['menu_selections']['DIGITS'])==0:
			return
		d=self.window.data['menu_selections']['DIGITS']
		b=self.menu_buttons['DIGITS']
		b.select_sub(d[str(b.sub_group['0'])])#key str(b.sub_group['0']) is the same for all items in 'DIGITS', so using '0'
		d=self.window.data['menu_selections']['REGRESSION']
		for caption in d:
			if not caption in ["['JOINED', 'JOINED LONG', 'disabled:JOINED']",
							   "['NORMAL', 'HTML', 'LATEX', 'RTF', 'INTERNAL']"]:
				self.menu_buttons['REGRESSION'].click_sub(caption,d[caption],False)
			
	def save(self):
		"Saves the table part of the screen"
		#Problem: don't save all if table is bigger than screen. possible soloution, a popup window containing only the filled part of the text widget
		t=self.tab.widget.text_box
		img=ImageGrab.grab((t.winfo_rootx(),t.winfo_rooty(),t.winfo_width(),t.winfo_height()))
		img.save(os.path.join(self.window.data['current path'],'screen.png'))
		
	def data_doesnt_exist(self):
		"tests if data exists"
		if not hasattr(self.widget.stored_output_data,'data'):
			return True
		if self.widget.stored_output_data.data is None:
			print("Data not found")
			return True
		return False

	def set_output_obj(self,ll, direction,main_msg):
		"sets the outputobject in the output" 
		self.output=output.output(ll, direction,main_msg)
		
		
	def get_digits(self):
		"returns the selected number of digits"
		b=self.menu_buttons['DIGITS']
		for i in b.buttons_sub:
			if b.buttons_sub[i]['fg']==fg_normal:
				try:
					return int(i)
				except:
					return i

	def get_active(self):
		"returns the active menu button"
		for i in self.menu_buttons:
			b=self.menu_buttons[i]
			if b.button_main['bg']==bg_selected:
				return i	
		
	#****UPDATE CALLS********
		
	def update_after_direction(self,direction,its):
		self.output.update_after_direction(direction,its)
		self.reg_table=self.output.reg_table()
		self.print()
		
	def update_after_linesearch(self,direction,ll,incr):
		self.output.update_after_linesearch(direction,ll,incr)
		if self.menu_buttons['DIAGNOSTICS'].button_main['bg']==bg_selected:
			self.statistics=self.output.statistics()
			self.widget.stored_output_data.statistics=self.statistics
		self.reg_table=self.output.reg_table()
		self.widget.stored_output_data.reg_table=self.reg_table		
		
		if self.menu_buttons['DISTRIBUTION CHARTS'].button_main["fg"]==fg_normal:
			self.charts.plot(ll)
		self.widget.stored_output_data.chart_images=self.charts.get_images_for_storage()#for storing the editor
		self.widget.stored_output_data.data=stored_data(ll,direction,self.reg_table)#for storing the editor		
		self.print()
		
	#****END UPDATE CALLS********
		

			
	#********MENU METHODS***********
		
	def print_regression(self):
		if self.reg_table is None:
			self.reg_table=self.widget.stored_output_data.reg_table
			if self.reg_table is None:return
		#for storing the editor
		n=self.get_digits()
		stacked=self.menu_buttons['REGRESSION'].buttons_sub["['FLAT', 'STACKED']"]['text']!='STACKED'
		if self.menu_buttons['REGRESSION'].buttons_sub[ "['( )', '[ ]', 'disabled:( )']"]['fg']==fg_disabled:
			bracket=''
		elif self.menu_buttons['REGRESSION'].buttons_sub[ "['( )', '[ ]', 'disabled:( )']"]['text']=='[ ]':
			bracket='['
		else:
			bracket='('
		fmt=self.menu_buttons['REGRESSION'].buttons_sub["['NORMAL', 'HTML', 'LATEX', 'RTF', 'INTERNAL']"]['text']
		joined=self.menu_buttons['REGRESSION'].buttons_sub["['JOINED', 'JOINED LONG', 'disabled:JOINED']"]
		if joined['fg']!=fg_normal:
			self.print_regression_single(n,stacked, bracket, fmt)
		else:
			self.join_tables(n, stacked,bracket, fmt,joined['text'])
			
	def show_scatter(self):
		if self.data_doesnt_exist():return
		d=self.widget.stored_output_data.data	
		#gui_scatter_charts.scatter_window(self,d.X_names,d.Y_name,d.X_st,d.Y_st,self.window.iconpath,self.tabs,700,1000)	
		if self.menu_buttons['SCATTER PLOTS'].buttons_sub['NORMALIZED']['fg']!=fg_disabled:
			if d.changed_since_last_scatter:
				self.scatter_norm.plot(d.X_names,d.Y_name,d.X_st,d.Y_st)
				d.changed_since_last_scatter=False
			self.scatter_norm.grid(column=0,row=1,sticky=tk.NSEW)
			self.scatter.grid_forget()
		else:
			if self.scatter.plotted==False:
				self.scatter.plot(d.X_names,d.Y_name,d.X,d.Y)
			self.scatter.grid(column=0,row=1,sticky=tk.NSEW)
			self.scatter_norm.grid_forget()
		self.tab.widget.grid_forget()

	def print_correl(self):
		if self.data_doesnt_exist():return
		d=self.widget.stored_output_data.data
		is_covar=self.menu_buttons[' CORREL '].buttons_sub["[' CORREL ', ' COVAR ']"]['text']==' COVAR '
		X,Y=d.X,d.Y
		if self.menu_buttons[' CORREL '].buttons_sub["NORMALIZED"]['fg']==fg_normal:
			X,Y=d.X_st,d.Y_st		
		if self.menu_buttons[' CORREL '].buttons_sub["['VARIABLES', 'ESTIMATES', 'RAW MOMENTS']"]['text']=='ESTIMATES':
			if d.hessian is None: return
			if is_covar:
				c=d.hessian
			else:
				n=len(d.hessian)
				s=np.maximum(np.diag(-d.hessian),0)**0.5
				s=s.reshape(n,1)*s.reshape(1,n)
				s[s==0]=d.hessian[s==0]
				c=-d.hessian/s
			names=d.args_names
		elif self.menu_buttons[' CORREL '].buttons_sub["['VARIABLES', 'ESTIMATES', 'RAW MOMENTS']"]['text']=='VARIABLES':
			c=st.correl_2dim(np.concatenate((Y,X),1),covar=is_covar)	
			names=d.all_names
		else:#RAW MOMENTS (remove?)
			V=np.concatenate((Y,X),1)
			c=np.dot(V.T,V)
			names=d.all_names
		correl=[['']+names]
		for i in range(len(c)):
			correl.append([names[i]]+list(c[i]))
		s=self.generate_table(correl)
		formatting={'single_line':[1],
					'double_line':[3],
					'bold':[2],
					'bold_col':2,
					'line_length':[len(c[0])]}		
		self.print_text(s,correl,formatting)
		
	def print_stats(self):
		if self.statistics is None:#in that case, this is an active output tab
			self.statistics=self.output.statistics()
			self.widget.stored_output_data.statistics=self.statistics			
			if self.statistics is None:return
		#for storing the editor
		n=self.get_digits()

		#tab_stops=self.reg_table.get_tab_stops(self.tab.widget.text_box.config()['font'][4])
		#self.tab.widget.text_box.config(tabs=tab_stops)	
		formatting={'bold_underline':[1]}
		tab_stops=('30',tk.LEFT,'60',tk.LEFT,'290',tk.RIGHT,'450',tk.NUMERIC)
		if self.menu_buttons['DIAGNOSTICS'].buttons_sub['SAMP.SIZE']['fg']==fg_normal:
			s="SAMPLE SIZE:\n\n"
			self.print_text(s+self.statistics.df_str,formatting=formatting,tab_stops=tab_stops)	
		elif self.menu_buttons['DIAGNOSTICS'].buttons_sub['TESTS']['fg']==fg_normal:
			s="DIAGNOSTICS:\n\n"
			s+=self.statistics.gen_mod_fit(n)
			s+=self.statistics.adf_str(n)
			self.print_text(s,formatting=formatting,tab_stops=tab_stops)
		else:
			s="ALL:\n\n"
			s+=self.statistics.df_str
			s+="\n  DIAGNOSTICS:"
			s+=self.statistics.gen_mod_fit(n)
			s+=self.statistics.adf_str(n)			
			self.print_text(s,formatting=formatting,tab_stops=tab_stops)	
			
	def print_descriptive_statistics(self,exp=False):
		if self.data_doesnt_exist():return
		d=self.widget.stored_output_data.data
		data=np.concatenate((d.Y,d.X),1)
		if self.menu_buttons['DESCRIPTIVES'].buttons_sub['EXP']['fg']!=fg_disabled:
			data=np.exp(data)
		elif self.menu_buttons['DESCRIPTIVES'].buttons_sub['LN']['fg']!=fg_disabled:
			data=np.log(data)
		n,k=data.shape
		stat=[['']+d.all_names]
		for operator,name in [
			(np.mean,'Mean:'),
			(np.median,'Median:'),
			(np.std,'S.D.:'),
			(np.min,'Minimum:'),
			(np.max,'Maximum:')]:
			stat.append([name]+list(operator(data,0)))
		stat_t=[]
		for i in range(len(stat[0])):
			stat_t.append([])
			for j in range(len(stat)):
				stat_t[i].append(stat[j][i])
		s=self.generate_table(stat_t)
		formatting={'single_line':[1],
					'double_line':[3],
					'bold':[2],
					'bold_col':2,
					'line_length':[len(stat_t[0])]}			
		self.print_text(s,stat_t,formatting)
		
		
	def print(self):
		for c in self.menu_buttons:
			if not (c in self.no_print_menus):
				b=self.menu_buttons[c]
				if b.button_main['bg']==bg_selected:
					b.command()
					return
				
			
	def show_dist_charts(self):
		b=self.menu_buttons['DISTRIBUTION CHARTS']
		if b.button_main['fg']==fg_disabled:
			self.charts.grid_forget()
		else:
			self.charts.grid(column=1,row=0,sticky=tk.NS)


	#********END MENU METHODS***********
		
	#******PRINTING METHODS*************
		
	def join_tables(self,digits,stacked,bracket,fmt,caption):
		"prints a table joined with previoius results"
		if self.widget.stored_output_data.data is None: return
		t=self.widget.stored_output_data.data.join_table
		if t is None:return
		s,X=t.make_table(stacked, bracket,digits,caption)	
		formatting=None
		n=len(t.names_v)
		if fmt=='NORMAL':
			formatting={'double_line':[4],
						'bold':[3],
						'bold_col':0,
						'line_length':[len(t)*(2-stacked)+3]
						}
			if stacked:
				formatting['10p']=[6+i*3 for i in range(len(t.names_v))]
				formatting['space_line']=[7+i*3 for i in range(len(t.names_v))]
				formatting['single_line']=[2,5+n*3,9+n*3]
			else:
				formatting['single_line']=[2,5+n,9+n]
		self.print_text(s,X,formatting)		
		
	def print_regression_single(self,digits,stacked,bracket,fmt):
		"prints a single regression"
		s,llength=self.reg_table.table(digits,bracket,fmt,stacked,
							   self.menu_buttons['REGRESSION'].buttons_sub['DIR']['fg']==fg_normal,
							   self.menu_buttons['REGRESSION'].buttons_sub['CNSTRNTS']['fg']==fg_normal)
		formatting=None
		if fmt=='NORMAL':
			formatting={'single_line':[5,9+self.reg_table.n_variables*3],
						'double_line':[7],
						'bold':[4,5,6],
						'bold_col':0,
						'line_length':[llength]
						}
			if stacked:
				formatting['10p']=[7+i*3 for i in range(self.reg_table.n_variables)]		
				formatting['space_line']=[8+i*3 for i in range(self.reg_table.n_variables)]		
		self.print_text(s,self.reg_table.X,formatting)
		
	def print_text(self,text,X=None,formatting=None,tab_stops="1c"):
		"prints the argument text"
		self.tab.widget.replace_all(text)	
		if not formatting is None:
			self.format_text_widget(formatting,text)	
		else:
			self.tab.widget.text_box.configure(font=("Courier", 10))	
		if not X is None:
			tab_stops=output.get_tab_stops(X,self.tab.widget.text_box.config()['font'][4])
		self.tab.widget.text_box.config(tabs=tab_stops)			
		self.scatter.grid_forget()
		self.scatter_norm.grid_forget()
		self.tab.widget.grid(column=0,row=1,sticky=tk.NSEW)
		

		
	def format_text_widget(self,formatting,text):
		"formats the output widget accoring to formatting"
		t=self.tab.widget.text_box
		t.configure(font=('Garamond', 12))
		if 'line_length' in formatting:
			self.add_lines(formatting['line_length'])
			for i in formatting['single_line']:
				t.image_create(f'{i}.0',image=self.txt_line)
			for i in formatting['double_line']:
				t.image_create(f'{i}.0',image=self.txt_double_line)	
		if 'bold' in formatting:
			for i in formatting['bold']:
				t.tag_add('bold_G',f'{i}.0' , f'{i}.end')
		if 'bold_underline' in formatting:
			for i in formatting['bold_underline']:
				t.tag_add('bold_G_underline',f'{i}.0' , f'{i}.end')		
		if '10p' in formatting:
			for i in formatting['10p']:
				t.tag_add('10p',f'{i}.0' , f'{i}.end')		
		if 'space_line' in formatting:
			for i in formatting['space_line']:
				t.tag_add('space_line',f'{i-1}.end' , f'{i+1}.0')
		if 'bold_col' in formatting:
			if formatting['bold_col']>0:
				for i in range(formatting['double_line'][-1]+1,text.count('\n')+1):
					ix0=f'{i}.0'
					for j in range(formatting['bold_col']):
						ix=t.search('\t',ix0,f'{i}.end')
						t.tag_add('bold_G',ix0 ,ix)
						ix0=f"{i}.{int(ix.split('.')[1])+1}"
						
	def generate_table(self,tbl):
		"generates a table as a text string from tbl"
		n=self.get_digits()
		for i in range(len(tbl)):
			for j in range(len(tbl[0])):
				try:
					tbl[i][j]=np.round(tbl[i][j],n)
				except:
					tbl[i][j]=tbl[i][j]
		s="\n"
		for i in range(len(tbl)):
			if i in [1]:
				s+='\n'			
			s+="\t"
			for j in range(len(tbl[0])):
				s+=f"{tbl[i][j]}\t"	
			s+="\n"
		return s	
						
		#******END PRINTING METHODS*************
		
		
class menu_button:
	"Class for the top level menu buttons"
	def __init__(self,output_tab,loc,caption_main,command,captions_sub,click_type,group_type_sub,sub_enabled):
		self.tab=output_tab	
		self.loc=loc
		self.button_frame=tk.Frame(self.tab.selection_frame,bg="white")
		if click_type is None:
			self.button_main=tk.Label(self.button_frame,text=caption_main, 
											   highlightthickness=0,bd=0, anchor=tk.E,bg=bg_normal,fg=fg_normal,font=self.tab.font)
		else:
			self.button_main=tk.Button(self.button_frame,text=caption_main,command=self.click, 
									   highlightthickness=0,bd=0, anchor=tk.E,bg=bg_normal,fg=fg_normal,font=self.tab.font)
		self.button_main.bind("<Motion>", self.mouseover)
		self.button_main.bind("<Leave>", self.button_hide_sub)
		self.button_main.grid(row=0,column=0)
		tk.Label(self.button_frame,text="|",bg=bg_normal,font=self.tab.font).grid(row=0,column=1)
		self.button_frame.grid(sticky='w',row=0,column=loc)	
		self.command=command
		self.captions_sub=captions_sub
		self.click_type=click_type
		self.group_type_sub=group_type_sub
		self.caption=caption_main
		self.sub_enabled=sub_enabled
		self.add_sub_menu()
		if not caption_main in self.tab.window.data['menu_selections']:
			self.tab.window.data['menu_selections'][caption_main]=dict()
		self.selections=self.tab.window.data['menu_selections'][caption_main]
		
	def select(self,withdraw=True):
		for i in self.tab.menu_buttons:
			if not i==self.caption and self.click_type=='group':
				self.tab.menu_buttons[i].alter_state(bg_normal,fg_normal)#reset all menus to unselected
			if withdraw:#If withdraw, then all button_frames are hidden
				self.tab.menu_buttons[i].button_box.withdraw()
		self.alter_state(bg_selected,bg_normal)	#set current menu to selected

		
	def toggle(self):
		if self.button_main['fg']==fg_normal:
			self.button_main.configure(fg=fg_disabled)
		else:
			self.button_main.configure(fg=fg_normal)			
		
	def click(self,runcommand=True):
		if self.click_type=='group':
			self.select()
		elif self.click_type=='toggle':
			self.toggle()
		if (self.command is None) or (self.click_type is None):
			return
		try:
			self.command()
		except Exception as e:
			traceback.print_exc()	
			
	def mouseover(self,event):
		if self.n_sub>0:
			self.show_sub()			
		
	def alter_state(self,bg,fg):
		self.button_main.configure(fg=fg)
		self.button_main.configure(bg=bg)
		
	def show_sub(self):
		x, y, cx, cy = self.button_main.bbox("insert")
		x += self.button_main.winfo_rootx()
		y += self.button_main.winfo_rooty()+self.button_frame.winfo_height()
		w=self.button_frame.winfo_width()
		h=(self.button_frame.winfo_height()+5)*self.n_sub
		self.button_box.geometry('%dx%d+%d+%d' % (150,h,x,y))
		for c in self.buttons_sub:
			self.buttons_sub[c].configure(widt=self.button_main.winfo_width())
		self.button_box.deiconify()	
		
	def hide_sub(self,event=None):
		w,h,x,y=split_geometry_string(self.button_box.geometry())
		if event.widget._name=='!toplevel' and event.y>0:
			self.button_box.withdraw()	
			
	def button_hide_sub(self,event):
		if event.y<self.button_main.winfo_height():
			self.button_box.withdraw()
		
	def click_sub(self,caption,i,execute=True):
		if type(self.captions_sub[i])==list:
			self.toggle_list(self.captions_sub[i],i)
			self.selections[str(caption)]=i
		elif self.sub_group[caption] is None:
			if self.group_type_sub==[None]:
				r=self.toggle_sub_select(caption)
			else:
				r=self.toggle_sub(caption)
			self.selections[caption]=r
		else:
			self.select_sub(caption)
			self.selections[str(self.sub_group[caption])]=caption
		if not execute:
			return
		if self.click_type=='group':
			self.select(False)
		if self.command is None:
			return
		try:
			self.command()
		except Exception as e:
			traceback.print_exc()
			
	def select_sub(self,caption):
		if self.buttons_sub[caption]['fg']==fg_disabled:
			for i in self.sub_group[caption]:
				self.buttons_sub[i].configure(fg=fg_disabled)
			self.buttons_sub[caption].configure(fg=fg_normal)
			
	def toggle_sub_select(self,caption):
		r=self.toggle_sub(caption)
		if self.buttons_sub[caption]['fg']==fg_normal:
			for i in self.buttons_sub:
				if not i==caption:
					self.buttons_sub[i].configure(fg=fg_disabled)
		return r
			
	def toggle_sub(self,caption):
		if self.buttons_sub[caption]['fg']==fg_disabled:
			self.buttons_sub[caption].configure(fg=fg_normal)
			return True
		else:
			self.buttons_sub[caption].configure(fg=fg_disabled)		
			return False
			
	def toggle_list(self,lst,i):
		caption=str(lst)
		if self.buttons_sub[caption]['fg']==fg_disabled:
			k=0
		else:
			current=self.buttons_sub[caption]['text']
			k=lst.index(current)+1
			if k==len(lst):
				k=0
		if 'disabled:' in lst[k]:
			self.buttons_sub[caption].configure(text=lst[k].split(':')[1])
			self.buttons_sub[caption].configure(fg=fg_disabled)
		else:
			self.buttons_sub[caption].configure(text=lst[k])
			self.buttons_sub[caption].configure(fg=fg_normal)
	
	def add_sub_menu(self):
		# creates a toplevel window
		self.button_box = tk.Toplevel(self.button_main,bg=bg_sub_menu)
		self.button_box.withdraw()
		self.button_box.bind("<Leave>", self.hide_sub)
		# Leaves only the label and removes the app window
		self.button_box.wm_overrideredirect(True)
		frm = tk.Frame(self.button_box, background=bg_sub_menu, relief='flat')
		if self.captions_sub is None:
			self.n_sub=0
		else:
			self.n_sub=len(self.captions_sub)
			
		self.buttons_sub={}
		self.captions_sub_str=[]
		fg=fg_disabled
		if self.sub_enabled=='all':
			fg=fg_normal
		for i in range(self.n_sub):
			c=str(self.captions_sub[i])
			self.captions_sub_str.append(c)
			if type(self.captions_sub[i])==list:
				cmd=eval(f"lambda: self.click_sub({c},{i})",{'self':self})
				text=self.captions_sub[i][0]
			else:
				cmd=eval(f"lambda: self.click_sub('{c}',{i})",{'self':self})
				text=c
			self.buttons_sub[c]=tk.Button(frm, text = text,command=cmd,
														font=self.tab.font, highlightthickness=0,bd=0,bg=bg_sub_menu,fg=fg,anchor="w")
			if (not self.sub_enabled is None) and i in self.sub_enabled:
				self.buttons_sub[c].configure(fg=fg_normal)
			else:
				if 'disabled:'==c[:9]:
					self.buttons_sub[c].configure(text=c[9:])
			self.buttons_sub[c].grid(row=i,column=0,sticky=tk.E)
		self.define_sub_grups()
		frm.grid(padx=10)

	def define_sub_grups(self):
		if self.captions_sub is None:
			return
		if self.group_type_sub=='all':
			self.group_type_sub=[self.captions_sub_str]
			self.sub_group={str(i):None for i in self.captions_sub_str}
			return
		self.sub_group={str(i):None for i in self.captions_sub_str}
		if not type(self.group_type_sub)==list:
			return
		g=[]
		n=len(self.group_type_sub)
		for i in range(n):
			g.append([])
			group=self.group_type_sub[i]
			if group is None:
				g[i]=None
			else:
				if not (type(group)==list or type(group)==range):
					raise RuntimeError("group_type_sub needs to be a list of lists or ranges")
				for j in group:
					g[i].append(self.captions_sub_str[j])
				for j in group:
					self.sub_group[self.captions_sub_str[j]]=g[i]
		self.group_type_sub=g
		

class stored_output:#for storing the editor
	def __init__(self):
		self.chart_images=None
		self.data=None
		self.reg_table=None
		self.statistics=None
		
		

class stored_data:
	def __init__(self,ll,direction,reg_table):
		N,T,k=ll.panel.X.shape
		panel=ll.panel
		self.X=panel.input.X
		self.Y=panel.input.Y
		if not hasattr(ll,'X_st'):
			ll.standardize()
		self.X_st=ll.X_st[panel.included[2]]
		self.Y_st=ll.Y_st[panel.included[2]]		
		self.X_names=panel.input.X_names
		self.Y_name=panel.input.Y_names
		self.args_names=panel.args.names_v
		self.all_names=panel.input.Y_names+list(panel.input.X_names)
		self.args=ll.args
		self.LL=ll.LL
		self.changed_since_last_scatter=True
		self.hessian=direction.H
		self.gradient=direction.g
		self.descr=panel.input.descr
		self.reg_stats=reg_table.d
		self.add_join_table(panel,ll)
		
	def add_join_table(self,panel,ll):
		if panel.input.join_table is None:
			self.join_table=None
			return
		elif len(panel.input.join_table)==0:
			panel.input.join_table.append(output.join_table(ll.args))
		elif type(panel.input.join_table[0])==str:
			if len(panel.input.join_table)>1:
				varnames=list(panel.input.join_table)
			else:
				s=panel.input.join_table
				varnames=s.replace(',','+').split('+')
			while len(panel.input.join_table):
				panel.input.join_table.pop()#removes all elements
			panel.input.join_table.append(output.join_table(ll.args,varnames))
		self.join_table=panel.input.join_table[0]
		self.join_table.update(ll,self.reg_stats,self.descr)
		
class bar(tk.Frame):
	def __init__(self,master,exe_tab):

		tk.Frame.__init__(self,master,background=bg_normal,height=75)
		self.tab=master
		self.exe_tab=exe_tab
		self.text_frame=tk.Frame(self,bg=bg_normal)
		self.text_frame_L=tk.Frame(self.text_frame,bg=bg_normal)
		self.text_frame_R=tk.Frame(self.text_frame,bg=bg_normal)
		self.text_frame_R.grid(row=0,column=1,sticky=tk.E)
		self.task_text=tk.StringVar(self)
		self.task_lbl=tk.Label(self.text_frame_R,textvariable=self.task_text,background=bg_normal,anchor='e', justify=tk.RIGHT,width=15)
		self.info_txt=tk.Text(self.text_frame_L,height=3,width=120,borderwidth = 0, highlightthickness = 0)
		self.info_txt.grid(row=0,column=0,sticky=tk.EW)
		self.info_txt.tag_configure('bold', font="Courier 10 bold")
		
		
		
		self.task_lbl.grid(row=0,column=0,sticky=tk.E)
		#self.text_frame_L.columnconfigure(0,weight=1)
		self.text_frame.columnconfigure(0,weight=1)
		
		self.columnconfigure(0,weight=1)
		
		self.text_frame_L.grid(row=0,column=0,sticky=tk.EW)
		self.text_frame.grid(row=0,column=0,sticky=tk.EW)
		self.progress=tk.Frame(self,background='#9cff9d',height=5,width=0)
		self.progress.grid(row=1,column=0,sticky=tk.W)
		
	def set_progress(self,percent=None,text="",task=''):
		total_width=self.winfo_width()
		if not percent==None:
			self.progress.config(width=int(total_width*percent))
			self.progress.grid()
		if len(task):
			self.set_task(task)
		if len(text):
			self.print(text)
		if not self.exe_tab is None:
			return self.exe_tab.isrunning
		else:
			return True
		
	def set_task(self,task):
		if task=='linesearch':
			self.progress.config(bg='#9cff9d')
			self.task_lbl.config(bg='#9cff9d')
			self.task_text.set('Linesearch')
			
		if task=='gradient':
			self.progress.config(bg='#8dd68e')
			self.task_lbl.config(bg='#8dd68e')
			self.task_text.set('Calculating gradient')	
			
		if task=='hessian':
			self.progress.config(bg='#82bd83')
			self.task_lbl.config(bg='#82bd83')
			self.task_text.set('Calculating hessian')	
			
		if task=='done':
			self.progress.config(bg='#9cff9d')
			self.task_lbl.config(bg='#9cff9d')
			self.task_text.set('Done')		
			
		if task=='err':
			self.progress.config(bg='red')
			self.task_lbl.config(bg='red')
			self.task_text.set('No convergence')			
			
		
		
	def print(self,text):
		n=max((int(self.info_txt.index('end').split('.')[0]),1))
		text='>'+text
		if len(self.info_txt.get('1.0','end')):
			text='\n'+text
		self.info_txt.insert('end',text)
		self.info_txt.tag_add('bold',f'{n}.0' , f'{n}.end')
		self.info_txt.tag_remove('bold','1.0' , f'{n-1}.end')
		self.info_txt.see('end')

				
		
		
		
class selection_frame(tk.Frame):
	def __init__(self, master,):
		tk.Frame.__init__(self,master,background=bg_normal)
		
	def grid(self,column,row):
		super().grid(column=column,row=row,sticky=tk.EW)

	
def split_geometry_string(s):
	s,x,y=s.split('+')
	w,h=s.split('x')
	return int(w),int(h),int(x),int(y)


		



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from tkinter import ttk
from gui import gui_charts
from gui import gui_data_objects
from gui import gui_options

	
class right_tab_widget:
	def __init__(self,window):
		self.win=window
		frm=tk.Frame(window.frm_right)
		self.sel_data=tk.StringVar(frm)
		lbl_desc=tk.Label(frm,text='Selected dataset:',background='white',font='Arial 9')
		lbl=tk.Label(frm,textvariable=self.sel_data,background='white',font='Arial 9 bold')
		
		self.tabs = ttk.Notebook(window.frm_right)          # Create Tab Control	
		
		self.optionset=gui_options.optionset(self.tabs,window)
		self.data_tree=gui_data_objects.data_objects(self.tabs,window,self)
		self.preferences,pref_frame=gui_options.add_preferences_tab(self.tabs,window)
		
		self.tabs.add(self.data_tree.main_frame, text='datasets')      # Add the tab
		self.tabs.add(self.optionset.main_frame, text='options')      # Add the tab	
		self.tabs.add(pref_frame, text='preferences')      # Add the tab		
		
		lbl_desc.grid(row=0,column=0)
		lbl.grid(row=0,column=1)
		frm.grid(row=0,column=0,sticky=tk.NW)
		self.tabs.grid(row=1,column=0,sticky=tk.NSEW)  # Pack to make visible	
		#self.tabs.bind("<<NotebookTabChanged>>", self.main_tab_pressed)
		self.sql_script=''
		
		
		
	def main_tabs_pressed(self,event):
		pass
		
		
		
		



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import tkinter as tk
from gui import gui_functions as guif
import functions as fu
from shutil import copyfile

class scatter_window(tk.Toplevel):
	def __init__(self, master,X_names,Y_names,X,Y,iconpath,tabs,height=400,width=1000):
		tk.Toplevel.__init__(self, master,height=height,width=width,)	
		self.title('Scatter charts')
		self.geometry('%sx%s' %(width,height))
		self.iconbitmap(iconpath)
		self.rowconfigure(0,weight=1)
		self.columnconfigure(0,weight=1)			
		
		self.main_frame=scatter_charts(self,tabs)
		self.main_frame.plot(X_names,Y_names,X,Y)
		self.main_frame.grid(row=0,column=0)
		
		
		self.transient(master) #set to be on top of the main window
		self.grab_set() #hijack all commands from the master (clicks on the main window are ignored)
		#master.wait_window(self) #pause anything on the main window until this one closes (optional)		
	
class scatter_charts(tk.Frame):
	def __init__(self, master,tabs):
		tk.Frame.__init__(self, master)
		self.rowconfigure(0,weight=1)
		self.plotted=False
		self.columnconfigure(0,weight=1)	
		self.rowconfigure(1)
		self.columnconfigure(1)	
		self.rowconfigure(2)
		self.n_cols=3
		self.col_height=250		
		self.subplot=tabs.subplot
		self.print_subplot=tabs.print_subplot
		self.yscrollbar = tk.Scrollbar(self)
		self.xscrollbar = tk.Scrollbar(self,orient = tk.HORIZONTAL)
		#self.btn_saveall = tk.Button(self, text='Save all', command=self.saveall,bg='white')
		#self.btn_saveall.grid(row=1,column=0)	
		self.canvas=tk.Canvas(self,yscrollcommand = self.yscrollbar.set,xscrollcommand = self.xscrollbar.set,
							  width=2000,height=1000,bg='white')
		self.wdgt_frame=tk.Frame(self.canvas,bg='white')
		
		self.yscrollbar.config(command = self.canvas.yview)	
		self.xscrollbar.config(command = self.canvas.xview)	
		
		self.yscrollbar.grid(row=0,column=1,sticky='ns')
		self.xscrollbar.grid(row=2,column=0,sticky='ew')
		self.canvas.grid(row=0,column=0,sticky=tk.NSEW)		
		
	def plot(self,X_names,Y_names,X,Y):
		self.X=X
		self.Y=Y		
		self.n_plots=X.shape[1]
		self.n_rows=int(self.n_plots/self.n_cols)
		self.n_rows+=(self.n_rows*self.n_cols<self.n_plots)
		self.canvas.configure(scrollregion=(0,0,1000,self.n_rows*self.col_height))
		self.wdgt_frame.configure(height=self.n_rows*self.col_height,width=1000)
		self.wdgt_frame.grid(row=0,column=0,sticky=tk.NSEW)
		self.plot_all(X_names,Y_names)	
		self.canvas.create_window(0,0,window=self.wdgt_frame,anchor='nw')
		self.plotted=True
		
		
	def resize(self,event=None):
		print(self.winfo_width())
		


	def plot_all(self,X_names,Y_names):
		self.charts=dict()
		for i in range(self.n_rows):
			self.wdgt_frame.rowconfigure(i,weight=1)	
		for i in range(self.n_cols):
			self.wdgt_frame.columnconfigure(i,weight=1)			
		for row in range(self.n_rows):
			for col in range(self.n_cols):
				i=row*self.n_cols+col
				if i>=self.n_plots:
					break
				self.charts[(row,col)]=self.plot_scatter(i,X_names,Y_names,bgframe=self.wdgt_frame)
		for i in self.charts:			
			self.charts[i].grid(row=i[0],column=i[1])	
					
	def plot_scatter(self,i,X_names,Y_names,subplot=None,f=None,bgframe=None):
		if subplot is None:
			subplot=self.subplot
		fgr,axs=subplot
		x=self.X[:,i]
		y=self.Y[:,0]
		
		axs.scatter(x,y, alpha=.1, s=10)
		axs.yaxis.label.set_text(Y_names[0])
		axs.xaxis.label.set_text(X_names[i])
		name=f'{X_names[i]}'
		axs.set_title(name)	
		
		if f is None:
			w=int(fgr.get_figwidth()*fgr.get_dpi())
			h=int(fgr.get_figheight()*fgr.get_dpi())
			chart=tk.Label(bgframe,width=w,height=h)
			path=fu.obtain_fname(f'./output/{name}.png')
			guif.display(chart,name,i,subplot,self.on_scatter_click,path)
			return chart
		
		else:
		
			guif.save(subplot,f)
					
	def on_scatter_click(self,event):
		f = tk.filedialog.asksaveasfile(mode='bw', defaultextension=".jpg",initialfile=f"{event.widget.name}.jpg")	
		if f is None:
			return
		ch=open(event.widget.path,'rb')
		f.write(ch.read())
		ch.close()
		f.close()
		
		
	def saveall(self):
		f = tk.filedialog.asksaveasfile(mode='bw', defaultextension=".jpg",initialfile="paneltime_scatter_plots.jpg")	
		fname=f.name
		f.close()
		for i in self.charts:
			f=open(gui.fix_fname(fname,i.name))
			self.plot_scatter(i.i, self.print_subplot,f)
			f.close()
			
	def on_closing(self):
		self.withdraw()
		
	def save(self,i):
		f = tk.filedialog.asksaveasfile(mode='bw', defaultextension=".jpg",initialfile=f"paneltime_scatter_plot_{i}.jpg")		
		self.plot_scatter(i, self.print_subplot,f)
		f.close()
					
		
class ResizingCanvas(tk.Canvas):
	def __init__(self,parent,**kwargs):
		tk.Canvas.__init__(self,parent,**kwargs)
		#print self.winfo_reqwidth(),self.winfo_reqheight() #>>>854, 404
		self.bind("<Configure>", self.on_resize)

	def on_resize(self,event):
		self.width = event.width   #>>>854
		self.height = event.height #>>>404
		self.config(width=self.width, height=self.height)


#!/usr/bin/env python
# -*- coding: utf-8 -*-


#THIS DOES NOT WORK!!!
import subprocess
import os
import pickle

def main():
	a=browser()
	a.send("Hello World")
	d=0

class browser:
	"""Creates a slave"""
	command = ["start",os.path.join(os.path.dirname(__file__),"io.html")]

	def __init__(self):
		"""Starts browser"""
		command='start '+os.path.join(os.path.dirname(__file__),"io.html")
		self.p = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,shell=True)
	
		self.t=transact(self.p.stdout,self.p.stdin)


	def send(self,msg):
		"Sends msg to the browser"
		self.t.send((msg,))          

	def receive(self):
		"waits to recive message from the borowser"
		answ=self.t.receive()
		return answ

	def kill(self):
		self.p.kill()





class transact():
	"""Local worker class"""
	def __init__(self,read, write):
		self.r=read
		self.w=write

	def send(self,msg):
		w=getattr(self.w,'buffer',self.w)
		pickle.dump(msg,w)
		w.flush()   

	def send_debug(self,msg,f):
		w=getattr(self.w,'buffer',self.w)
		write(f,str(w))
		pickle.dump(msg,w)
		w.flush()   	

	def receive(self):
		r=getattr(self.r,'buffer',self.r)
		u= pickle.Unpickler(r)
		try:
			return u.load()
		except EOFError as e:
			if e.args[0]=='Ran out of input':
				raise RuntimeError("""An error occured in one of the spawned sub-processes. 
Check the output in "slave_errors.txt' in your working directory or 
run without multiprocessing\n %s""" %(datetime.datetime.now()))
			else:
				raise RuntimeError('EOFError:'+e.args[0])
			
			
main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import time
import statproc as stat
import functions as fu



def dd_func_lags_mult(panel,ll,g,AMAL,de_xi,de_zeta,vname1,vname2,transpose=False, de_zeta_u=None):
	#de_xi is "N x T x m", de_zeta is "N x T x k" and L is "T x T"
	
	if de_xi is None or de_zeta is None:
		return None,None	
	(N,T,m)=de_xi.shape
	(N,T,k)=de_zeta.shape
	DLL_e=g.DLL_e.reshape(N,T,1,1)
	u_calc=False
	if de_zeta_u is None:
		de_zeta_u=de_zeta#for error beta-rho covariance, the u derivative must be used
	#ARIMA:
	if not AMAL is None:
		de2_zeta_xi=fu.dot(AMAL,de_zeta_u,False)#"T x N x s x m"
		if transpose:#only happens if lags==k
			de2_zeta_xi=de2_zeta_xi+np.swapaxes(de2_zeta_xi,2,3)#adds the transpose
		de2_zeta_xi_RE=ll.re_obj.ddRE(de2_zeta_xi,de_xi,de_zeta,ll.e,vname1,vname2)
	else:
		de2_zeta_xi=0
		de2_zeta_xi_RE=ll.re_obj.ddRE(None,de_xi,de_zeta,ll.e,vname1,vname2)
		if de2_zeta_xi_RE is None:
			de2_zeta_xi_RE=None
	if not de2_zeta_xi_RE is None:	
		de2_zeta_xi_RE = de2_zeta_xi_RE * DLL_e
		de2_zeta_xi_RE = np.sum(np.sum(de2_zeta_xi_RE,0),0)#and sum it

	#GARCH: 
	if panel.m>0:
		h_e_de2_zeta_xi = de2_zeta_xi * ll.h_e_val.reshape(N,T,1,1)
		h_2e_dezeta_dexi = ll.h_2e_val.reshape(N,T,1,1) * de_xi.reshape((N,T,m,1)) * de_zeta.reshape((N,T,1,k))

		d2lnv_zeta_xi = (h_e_de2_zeta_xi + h_2e_dezeta_dexi)
		
		if panel.N>1:
			d_mu = ll.args.args_d['mu'] * panel.mean(d2lnv_zeta_xi,1)
			d_mu = d_mu.reshape((N,1,m,k)) * panel.included[4]
		else:
			d_mu=0
		
		
		d2lnv_zeta_xi = fu.dot(ll.GAR_1MA, d2lnv_zeta_xi)
		
		d2lnv_zeta_xi = d2lnv_zeta_xi + d_mu
		
		d2lnv_zeta_xi=np.sum(np.sum(d2lnv_zeta_xi*g.dLL_lnv.reshape((N,T,1,1)),0),0)
	else:
		d2lnv_zeta_xi=None

	return d2lnv_zeta_xi,de2_zeta_xi_RE

def dd_func_lags(panel,ll,L,d,dLL,addavg=0, transpose=False):
	#d is "N x T x m" and L is "k x T x T"
	if panel.m==0:
		return None
	if d is None:
		return None		
	(N,T,m)=d.shape
	if L is None:
		x=0
	elif len(L)==0:
		return None
	elif len(L.shape)==3:
		x=fu.dot(L,d,False)#"T x N x k x m"
		if transpose:#only happens if lags==k
			x=x+np.swapaxes(x,2,3)#adds the transpose
	elif len(L.shape)==2:
		x=fu.dot(L,d).reshape(N,T,1,m)
	if addavg:
		addavg=(addavg*panel.mean(d,1)).reshape(N,1,1,m)
		x=x+addavg
	dLL=dLL.reshape((N,T,1,1))
	return np.sum(np.sum(dLL*x,1),0)#and sum it	


def add(iterable,ignore=False):
	"""Sums iterable. If ignore=True all elements except those that are None are added. If ignore=False, None is returned if any element is None. """
	x=None
	for i in iterable:
		if not i is None:
			if x is None:
				x=i
			else:
				x=x+i
		else:
			if not ignore:
				return None
	return x

def prod(iterable,ignore=False):
	"""Takes the product sum of iterable. If ignore=True all elements except those that are None are multiplied. 
	If ignore=False, None is returned if any element is None. """
	x=None
	for i in iterable:
		if not i is None:
			if x is None:
				x=i
			else:
				x=x*i
		else:
			if not ignore:
				return None
	return x

def fillmatr(X,max_T):
	k=len(X[0])
	z=np.zeros((max_T-len(X),k))
	X=np.concatenate((X,z),0)
	return X

def concat_matrix(block_matrix):
	m=[]
	for i in range(len(block_matrix)):
		r=block_matrix[i]
		C=[]
		for j in range(len(r)):
			if not r[j] is None:
				C.append(r[j])
		if len(C):
			m.append(np.concatenate(C,1))
	m=np.concatenate(m,0)
	return m

def concat_marray(matrix_array):
	arr=[]
	for i in matrix_array:
		if not i is None:
			arr.append(i)
	arr=np.concatenate(arr,2)
	return arr




		
def dd_func(d2LL_de2,d2LL_dln_de,d2LL_dln2,de_dh,de_dg,dln_dh,dln_dg,dLL_de2_dh_dg,dLL_dln2_dh_dg):
	a=[]
	a.append(dd_func_mult(de_dh,d2LL_de2,de_dg))

	a.append(dd_func_mult(de_dh,d2LL_dln_de,dln_dg))
	a.append(dd_func_mult(dln_dh,d2LL_dln_de,de_dg))

	a.append(dd_func_mult(dln_dh,d2LL_dln2,dln_dg))

	a.append(dLL_de2_dh_dg)
	a.append(dLL_dln2_dh_dg)
	return add(a,True)

def dd_func_mult(d0,mult,d1):
	#d0 is N x T x k and d1 is N x T x m
	if d0 is None or d1 is None or mult is None:
		return None
	(N,T,k)=d0.shape
	(N,T,m)=d1.shape
	if np.any(np.isnan(d0)) or np.any(np.isnan(d1)):
		x=np.empty((k,m))
		x[:]=np.nan
		return x
	d0=d0*mult
	d0=np.reshape(d0,(N,T,k,1))
	d1=np.reshape(d1,(N,T,1,m))
	try:
		x=np.sum(np.sum(d0*d1,0),0)#->k x m 
	except RuntimeWarning as e:
		if e.args[0]=='overflow encountered in multiply':
			d0=np.minimum(np.maximum(d0,-1e+100),1e+100)
			d1=np.minimum(np.maximum(d1,-1e+100),1e+100)
			x=np.sum(np.sum(d0*d1,0),0)#->k x m 
		else:
			raise RuntimeWarning(e)
	return x


def ARMA_product(m,k):
	a=[]

	for i in range(k):
		a.append(roll(m,-i-1,1))
	return np.array(a)


def sandwich(H,G,lags=3,ret_hessin=False):
	H=H*1
	sel=[i for i in range(len(H))]
	H[sel,sel]=H[sel,sel]+(H[sel,sel]==0)*1e-15
	hessin=np.linalg.inv(-H)
	V=stat.newey_west_wghts(XErr=G)
	hessinV=fu.dot(hessin,V)
	sandw=fu.dot(hessinV,hessin)
	if ret_hessin:
		return sandw,hessin
	return sandw


	
def differenciate(X,diff,has_intercept):
	for i in range(diff):
		X=X-np.roll(X,1,0)
	X=X[diff:]
	if has_intercept:
		X[:,0]=1
	return X


def roll(a,shift,axis=0,empty_val=0):
	"""For shift>0 (shift<0) this function shifts the shift up (down) by deleting the top (bottom)
	shift and replacing the new botom (top) shift with empty_val"""

	if shift==0:
		return a
	if type(a)==list:
		a=np.a(a)
	s=a.shape

	ret=np.roll(a,shift,axis)
	v=[slice(None)]*len(s)
	if shift>0:
		v[axis]=slice(0,shift)
	else:
		n=s[axis]
		v[axis]=slice(n+shift,n)
		ret[n+shift:]=empty_val
	ret[v]=empty_val

	if False:#for debugging
		arr2=a*1
		arr=a*1
		if len(s)==2:
			T,k=s
			fill=np.ones((abs(shift),k),dtype=arr2.dtype)*empty_val		
			if shift<0:
				ret2= np.append(fill,arr2[0:T+shift],0)
			else:
				ret2= np.append(arr2[shift:],fill,0)		
		elif len(s)==3:
			N,T,k=s
			fill=np.ones((N,abs(shift),k),dtype=arr2.dtype)*empty_val
			if shift<0:
				ret2= np.append(fill,arr2[:,0:T+shift],1)
			else:
				ret2= np.append(arr2[:,shift:],fill,1)		
		elif len(s)==1:
			T=s[0]
			fill=np.ones(abs(shift),dtype=arr2.dtype)*empty_val
			if shift<0:
				ret2= np.append(fill,arr2[0:T+shift],0)
			else:
				ret2= np.append(arr2[shift:],fill,0)	

		if not np.all(ret==ret2):
			raise RuntimeError('Check that the calling procedure has specified the "axis" argument')
	return ret




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import IPython
import webbrowser
import output
import os

WEB_PAGE='paneltime.html'



def get_channel(window,exe_tab):
	if not window is None:#tkinter gui
		return tk_widget(window,exe_tab)
	try:
		n=IPython.get_ipython().__class__.__name__
		if n=='ZMQInteractiveShell':
			return web_output(True)
	except:
		pass
	try:
		return web_output(False)
	except:
		pass
	return console()
		
class web_output:
	def __init__(self,Jupyter):
		self.Jupyter=Jupyter
		if not Jupyter:
			save_html(get_web_page('None', 'None', 'None', '', True))
			webbrowser.open(WEB_PAGE, new = 2)
			
			
		
	def set_progress(self,percent=None,text="",task=''):
		return True
		
	def set_output_obj(self,ll, direction,main_msg):
		"sets the outputobject in the output" 
		self.output=output.output(ll, direction,main_msg)
		
	def update_after_direction(self,direction,its):
		if not hasattr(direction,'ll'):
			return	
		self.its=its
		self.output.update_after_direction(direction,its)
		self.reg_table=self.output.reg_table()
		tbl,llength=self.reg_table.table(4,'(','HTML',True,
							   show_direction=True,
							   show_constraints=True)		
		web_page=get_web_page(direction.ll.LL, 
							  direction.ll.args.args_v, 
							  direction.dx_norm,
							  tbl,
							  self.Jupyter==False)
		if self.Jupyter:
			IPython.display.clear_output(wait=True)
			display(IPython.display.HTML(web_page))
		else:
			save_html(web_page)
		
	def update_after_linesearch(self,direction,ll,incr):
		if not hasattr(direction,'ll'):
			return			
		self.output.update_after_linesearch(direction,ll,incr)
		self.reg_table=self.output.reg_table()
		tbl,llength=self.reg_table.table(4,'(','HTML',True,
							   show_direction=True,
							   show_constraints=True)		
		web_page=get_web_page(ll.LL, 
							  ll.args.args_v, 
							  direction.dx_norm,
							  tbl,
							  self.Jupyter==False)
		if self.Jupyter:
			IPython.display.clear_output(wait=True)
			display(IPython.display.HTML(web_page))
		else:
			save_html(web_page)

		
class console:
	def __init__(self):
		pass
		
	def set_output_obj(self,ll, direction,msg_main):
		pass
		
	def update_after_direction(self,direction,its):
		pass
		
	def update_after_linesearch(self,direction,ll,incr):
		pass
				
class tk_widget:
	def __init__(self,window,exe_tab):
		self.tab=window.main_tabs._tabs.add_output(exe_tab)
		self.set_progress=self.tab.progress_bar.set_progress

		
	def set_output_obj(self,ll, direction,msg_main):
		self.tab.set_output_obj(ll, direction,msg_main)
		
	def update_after_direction(self,direction,its):
		self.tab.update_after_direction(direction,its)
		
	def update_after_linesearch(self,direction,ll,incr):
		self.tab.update_after_linesearch(direction,ll,incr)
		






def get_web_page(LL, args, direction,tbl,auto_update):
	au_str=''
	if auto_update:
		au_str="""<meta http-equiv="refresh" content="1" >"""
	img_str=''
	if os.path.isfile('img/chart0.png'):
		img_str=("""<img src="img/chart0.png"><br>\n"""
				"""<img src="img/chart1.png"><br>\n"""
				"""<img src="img/chart2.png">""")
	return f"""
<meta charset="UTF-8">
{au_str}
<head>
<title>paneltime output</title>
</head>
<style>
p {{
  margin-left: 60px;
  max-width: 980px;
  font-family: "verdana";
  text-align: left;
  color:#063f5c;
  font-size: 12;
}}
h1 {{
  margin-left: 20px;
  max-width: 980px;
  font-family: "verdana";
  text-align: left;
  color:black;
  font-size: 16;
}}
</style>
<body>
<div style='position:absolute;float:right;top:0;right:0'>
{img_str}
</div>
{tbl}
</body>
</html> """	


def save_html(htm_str):
	f = open(WEB_PAGE, "w")
	f.write(htm_str)
	f.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
import re
		
	
def edit_exe_script(window,dataset,editor=None):
	#Finds and edits an exe editor after click
	#to be removed when fixing this functionality
	editor=dataset.get_exe_editor(window.main_tabs,new_name='script')
	#original script
	if editor is None:
		editor=dataset.get_exe_editor(window.main_tabs,False)
		if dataset is None:
			return
		editor=dataset.get_exe_editor(window.main_tabs,False)
	script=editor.widget.get('1.0',tk.END)
	#insert execution part:
	exe_script=dataset.generate_exe_script(window.right_tabs.data_tree)
	m=re.search(r"execute\(([\s\S]*?)\)", script)
	if m is None:
		script=script+'\n\n'+exe_script
	else:
		script=script[:m.start()]+exe_script+script[m.end():]	
	imprt='from paneltime import *'
	if not imprt in ' '.join(script.split()):
		script=imprt+'\n'+script
	editor.widget.replace_all(script)
	window.data.save()
	return script
	
def edit_options_script(options):
	editor=options.dataset.get_exe_editor(options.win.main_tabs,False).widget
	script=editor.get('1.0',tk.END)
	#insert execution part:
	optns,search_patterns=options.get_script()
	m=re.search(r"execute\(([\s\S]*?)\)", script)
	if m is None:
		script=edit_exe_script(options.win,options.dataset,editor)
	end=re.search(r"execute\(([\s\S]*?)\)", script).start()
	if script[end-2:end]!='\n\n':
		script=script[:end]+'\n\n'+script[end:]
	for i in range(len(optns)):
		m=re.search(search_patterns[i], script[:end])
		if m is None:
			script=script[:end-2]+'\n'+optns[i]+ script[end-2:]
		else:
			script=script[:m.start()]+optns[i]+script[m.end():]		
	editor.replace_all(script)
	options.win.data.save()


def get_start_end(lines,identstr1,identstr2):
	start=-1
	end=-1
	for i in range(len(lines)):
		if lines[i]==identstr1[:-1] and start==-1:
			start=i
		elif lines[i]==identstr2[:-1] and start>-1:
			end=i
			break
	return start,end
		
def nicify(string):
	repl_list= [['\n\n\n','\n\n'],['\n\r\n','\n\n'],['\n\n\r','\n\n'],['\r\n\n','\n\n'],['  ',' ']]
	for s,r in repl_list:	
		while True:
			if s in string:
				string=string.replace(s,r)
			else:
				break
	return string



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from tkinter import ttk
from gui import gui_scrolltext
from gui import gui_output_tab
import functions as fu
import os
from tkinter import filedialog
import types
import numpy as np
from multiprocessing import pool
import paneltime as pt
import traceback
from matplotlib import pyplot  as plt
import tempstore




ADD_EDITOR_NAME='...'
DEFAULT_GREY='#e6e6e6'

class main_tabs(ttk.Notebook):
	def __init__(self,window):
		self.win=window
		ttk.Notebook.__init__(self,window.frm_left) 
		self.last_active_tab=None
		self.output_obj=None
		self.grid(column=0,row=0,sticky=tk.NSEW)  # Pack to make visible	
		self._tabs=tabs(self)
		self.isdeleting=False	
		self.bind("<<NotebookTabChanged>>", self.main_tab_pressed)
			
	def main_tab_pressed(self,event):	
		s=self.current_editor(0)
		if s==ADD_EDITOR_NAME:
			self._tabs.get_new_editor()
		else:
			self._tabs.selection_change(self.select())
			
	def insert_current_editor(self,chars):
		tb=self.current_editor(2)
		tb.write(chars)
		tb.text_box.focus()
		
	def current_editor(self,return_obj=1):
		selection = self.select()
		if return_obj==1:
			return str(self._tabs[selection].frame)#returns tab object
		elif return_obj==2:
			return self._tabs[selection].widget#returns widget
		return self.tab(selection, "text")

		
	def selected_tab_text(self):
		tb=self.current_editor(2)
		text=tb.get_all()	
		return text
	

		
		
class tab:
	def __init__(self,tabs,frame,widget,name,top_text='',top_color=DEFAULT_GREY,path=None,skip_buttons=False):
		self.frame = frame
		self.position=tabs.count
		self.isrunning=False
		self.tabs=tabs
		self.name=name
		self.path=path	
		self.top_color=top_color
		self.frame.rowconfigure(0)
		self.frame.rowconfigure(1, weight=1)
		self.frame.columnconfigure(0, weight=1)	
		self.notebook=tabs.notebook
		self.locals=dict()
		self.globals={'window':self.notebook.win,'tab':self,'data':self.notebook.win.right_tabs.data_tree.datasets}		
		self.add_buttons(name,top_color,top_text,skip_buttons)
		self.notebook.add(self.frame,text=name)
		self.widget = widget	
		self.locals=dict()
		if not widget is None:
			widget.grid(row=1, column=0,sticky=tk.NSEW)

		
	def add_buttons(self,name,top_color,top_text,skip_buttons):
		size=22
		self.button_frame=tk.Frame(self.frame,height=size,background=top_color)
		self.button_frame_L=tk.Frame(self.button_frame,height=size,background=top_color)
		self.button_frame_R=tk.Frame(self.button_frame,height=size,background=top_color)
		self.top_text=tk.StringVar(value=top_text,master=self.button_frame)
		self.button_frame_label_M=tk.Label(self.button_frame,textvariable=self.top_text,background=top_color)

		self.display_name=tk.StringVar(self.button_frame)
		self.display_name.set(name)
		self.name_box=tk.Entry(self.button_frame_L,width=50,textvariable=self.display_name)
		self.name_box.bind('<KeyRelease>', self.tab_name_edited)
		self.button_img=dict()		
		self.button_img['save']= tk.PhotoImage(file =  fu.join(os.path.dirname(__file__),['img','save.png']),master=self.button_frame)
		self.button_save=tk.Button(self.button_frame_L, image = self.button_img['save'],command=self.save, 
								   highlightthickness=0,bd=0,height=size, anchor=tk.E,background=top_color)
		

		
		self.button_img['delete']= tk.PhotoImage(file =  fu.join(os.path.dirname(__file__),['img','delete_small.png']),master=self.button_frame)
		self.button_delete=tk.Button(self.button_frame_R, image = self.button_img['delete'],command=self.delete, 
								   highlightthickness=0,bd=0,height=15, anchor=tk.E,background=top_color)
		
		
		self.button_frame.grid(row=0,sticky=tk.EW)
		self.button_frame.columnconfigure(0,weight=1)
		self.button_frame.columnconfigure(1,weight=5)
		self.button_frame.columnconfigure(2,weight=1)
		self.button_frame_L.grid(row=0,column=0,sticky=tk.W)
		self.button_frame_label_M.grid(row=0,column=1,sticky=tk.EW)
		self.button_frame_R.grid(row=0,column=2,sticky=tk.E)
		
		self.name_box.grid(row=0,column=0)
		self.button_delete.grid(row=0,column=2)
		if skip_buttons:
			return
		self.button_img['run']= tk.PhotoImage(file =  fu.join(os.path.dirname(__file__),['img','run.png']),master=self.button_frame)
		self.button_run=tk.Button(self.button_frame_L, image = self.button_img['run'],command=self.run, 
								   highlightthickness=0,bd=0,height=size, anchor=tk.E,background=top_color)		
		
		self.button_img['stop']= tk.PhotoImage(file =  fu.join(os.path.dirname(__file__),['img','stop.png']),master=self.button_frame)
		self.button_stop=tk.Button(self.button_frame_L, image = self.button_img['stop'],command=self.stop, 
								   highlightthickness=0,bd=0,height=size, anchor=tk.E,background=top_color)				
		
		
		self.button_save.grid(row=0,column=1)
		self.button_run.grid(row=0,column=2)
		self.button_stop.grid(row=0,column=3)
		
	def stop(self):
		self.tabs.notebook.win.mc.master.quit()
		self.tabs.notebook.win.mc=None
		if hasattr(self,'exe_tab'):
			self.exe_tab.isrunning=False
			self.exe_tab.pool.close()
		else:
			self.isrunning=False
			self.pool.close()	
		
	
	def run(self):
		win=self.notebook.win
		win.data.save()
		self.pool = pool.ThreadPool(processes=1)
		text=self.widget.get_all()
		self.globals['exe_tab']=self
		self.isrunning=True
		self.process=self.pool.apply_async(self.exec, (text,))

	def exec(self,source):
		try:
			exec(source,self.globals,self.locals)
		except Exception as e:
			print("""
The following error occured in you script:
------------------------------------------
""")
			traceback.print_exc()
	
	def run_enable(self,event):
		pass
		
	def tab_name_edited(self,event=None):
		new_name=self.display_name.get()
		if new_name in self.tabs.names:
			if not self.frame==self.tabs.names[new_name].frame:
				print('Name all ready exist')
				return
		self_name=self.name
		if (not self_name in self.tabs.names) or (new_name==self_name):
			return
		self.tabs.names[new_name]=self.tabs.names[self_name]
		self.tabs.names.pop(self_name)
		self.name=new_name
		self.notebook.tab(self.frame,text=new_name)
		
	def save(self):
		p=self.notebook.win.data['current path']
		initname=self.display_name.get().replace('.txt','')+'.txt'
		filename = filedialog.asksaveasfilename(initialdir=os.path.join(p,initname),title="Save",
			filetypes = (("text", "*.txt"),("Rich Text Format", "*.rtf"),("Python file", "*.py")), defaultextension=True)
		if not filename: 
			return
		self.path=filename
		p,f=os.path.split(filename)
		self.notebook.win.data['current path']=p
		txt=self.widget.get_all()
		file = open(filename,'w')
		file.write(txt)
		file.close()	
		self.display_name.set(initname)
	
	def delete(self):
		self.notebook.isdeleting=True
		self.tabs.pop(self.frame)
		
	
	
class tabs(dict):
	def __init__(self,notebook):
		dict.__init__(self)
		self.count=0
		self.names=dict()
		self.notebook=notebook
		self.sel_list=[]
		self.subplot=plt.subplots(1,figsize=(4,2.5),dpi=75)
		self.print_subplot=plt.subplots(1,figsize=(6,3.25),dpi=200)	
		self.add_tab = tk.Frame(self)
		self.add(self.add_tab,None,ADD_EDITOR_NAME)		
		self.load_all_from_temp()
	
	def add(self,frame,widget=None,name='',top_text='',top_color=DEFAULT_GREY,path=None,skip_buttons=False):
		
		tab_item=tab(self,frame, widget, name,top_text,top_color,path,skip_buttons=skip_buttons)
		dict.__setitem__(self,str(frame), tab_item)
		self.names[name]=self[frame]
		self.count+=1
		self.sel_list.append(str(frame))
		return self[frame]	
	
	def save_all_in_temp(self):
		win=self.notebook.win
		win.data['editor_data']=dict()
		for i in self:
			t=self[i]
			try:
				text=t.widget.get('1.0',tk.END)
			except:
				text=''
			try:			
				output_data=t.frame.widget.stored_output_data
			except:
				output_data=None
			if not t.name==ADD_EDITOR_NAME:
				win.data['editor_data'][i]=(t.name,text,t.top_text.get(),t.top_color,t.path,output_data)
				
	def load_all_from_temp(self):
		editor_data=self.notebook.win.data.get('editor_data')
		used_imgs=self.get_image_refs(editor_data)
		self.img_tmp=tempstore.temp_image_manager(used_imgs)	
		if editor_data is None:
			return
		n=0
		for i in editor_data:
			try:
				name,text, top_text, top_color,path,output_data=editor_data[i]
				if not (name==ADD_EDITOR_NAME or (name[:6]=='script' and text=='')):
					if output_data is None:
						self.add_editor(name,text,top_text=top_text,top_color=top_color,path=path).frame.focus()
					else:
						self.add_output(name,output_data=output_data)
					n+=1
			except:
				print(f'Could not add tab {i}')
		if n==0:
			self.get_new_editor().frame.focus()	
			
	def get_image_refs(self,editor_data):
		if editor_data is None:
			return []
		used_imgs=[]
		for i in editor_data:
			try:
				name,text, top_text, top_color,path,output_data=editor_data[i]
				if not output_data is None:
					for imgpath,name in output_data.chart_images:
						used_imgs.append(imgpath)
			except:
				pass
		return used_imgs
						
	def get_new_editor(self):
		return self.add_editor('script.py',top_text="editor",top_color='#fcdbd9')
			
	def add_editor(self,name=None,text=None,format_text=True,top_text='',top_color=DEFAULT_GREY,path=None):
		name=self.gen_name(name)
		frame=tk.Frame(self.notebook)
		widget= gui_scrolltext.ScrollText(frame,text=text,format_text=format_text,name=name,window=self.notebook.win)
		tab=self.add(frame,widget,name,top_text,top_color,path)
		frame.tab=tab
		self.notebook.select(frame)
		self.notebook.insert('end',self.add_tab)	
		self.remove_redundant_script(name)
		return tab
	
	def add_output(self,exe_tab=None,name='regression',output_data=None):
		output_tab=gui_output_tab.output_tab(self.notebook.win,exe_tab,name,self,self.notebook,output_data)
		#(added to tabs class in gui_output_tab.output_tab)
		return output_tab
	
	def __getitem__(self,k):
		k=str(k)
		ret=dict.__getitem__(self,k)
		return ret
		
	def __setitem__(self,k,v):
		raise RuntimeError("Can't assign values to the tabs object. Use the add method to create a new tab")
		
	def pop(self, k, d=None):
		if self[k].name==ADD_EDITOR_NAME:
			return
		add_new=False
		if self.count<=2:
			add_new=True
		self.select_valid_tab(k)		
		k=str(k)
		self.sel_list.pop(self.sel_list.index(k))
		try:
			self.names.pop(self[k].name)
		except:
			print(f"Could not delete {self[k].name} from {self.names}")
		self.notebook.forget(k)
		dict.pop(self,k, d)
		self.count-=1
		if add_new:
			self.get_new_editor()
			self.count+=1
		
	def selection_change(self,selection):
		i=self.sel_list.index(selection)
		self.sel_list.append(self.sel_list.pop(i))

	def select_valid_tab(self,current_tab):
		sel_list=self.sel_list
		for i in range(len(sel_list)):
			f=self[sel_list[-1-i]]
			if f.frame!=current_tab and f.name!='...':
				self.notebook.select(sel_list[-1-i])
				break
						
	def gen_name(self,core):
		if core is None:
			core='script'
		if 	not core in self.names:
			return core
		i=0
		n=list(core.split('.'))
		if len(n)==1:
			while True:
				i+=1
				name=f"{core} {i}"
				if 	not name in self.names:
					return name	
		else:
			while True:
				i+=1
				name=f"{'.'.join(n[:-1])} {i}.{n[-1]}"
				if 	not name in self.names:
					return name				
		
	def remove_redundant_script(self,name):
		if name[:6]=='script':
			return
		poptabs=[]
		for i in self:
			try:
				if self[i].name[:6]=='script' and self[i].widget.get_all().replace('\n','')=='':
					poptabs.append(i)
			except:
				pass
		for i in poptabs:
			self.pop(i)
	

		


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module contains statistical procedures

from scipy import stats as scstats
from scipy import special as sc
import calculus_functions as cf
import numpy as np
import loglikelihood as logl
import random_effects


def var_decomposition(XXNorm=None,X=None,concat=False):
	"""Variance decomposition. Returns the matrix of condition indexes for each factor (rows) and each variable
	(columns). Calculates the normalized sum of squares using square_and_norm if XXNorm is not supplied"""
	if XXNorm is None:
		XXNorm=square_and_norm(X)
	ub=len(XXNorm)     
	d,EVec=np.linalg.eig(XXNorm)
	if np.any(np.round(d.imag,15)!=len(d)*[0]):
		pass
		#print( "non-real XX matrix")
		
	d=d.real
	EVec=EVec.real
	d=np.abs(d)**0.5+1e-100
	MaxEv=np.max(d)  
	fi=np.abs(EVec*EVec/((d*d).reshape((1,ub))+1E-200))
	fiTot=np.sum(fi,1)
	pi=fi/fiTot.reshape((ub,1))
	pi=pi.T
	CondIx=MaxEv/d
	ind=np.argsort(CondIx)
	pi=pi[ind]
	CondIx=CondIx[ind]
	CondIx=CondIx.reshape((len(CondIx),1))
	if concat:
		return np.concatenate((CondIx,pi),1)
	else:
		return CondIx,pi

def square_and_norm(X):
	"""Squares X, and normalize to unit lenght.
	Similar to a correlation matrix, except the
	means are not subtracted"""
	N,T,k=X.shape
	Sumsq=np.sqrt(np.sum(np.sum(X**2,0),0))
	Sumsq.resize((k,1))
	Sumsq=Sumsq*Sumsq.T
	norm=cf.dot(X,X)/(Sumsq+1e-200)
	return norm

def singular_elim(panel,X):
	"""Eliminates variables that cause singularity"""
	N,T,k=X.shape
	r=np.arange(k)
	ci_threshold=50
	keep,XXCorrel=find_singulars(panel,X) 
	XXNorm=square_and_norm(X)
	cond_ix,pi=var_decomposition(XXNorm)
	if max(cond_ix)<ci_threshold:
		return keep,cond_ix
	for cix in range(1,len(cond_ix)):
		if (np.sum(pi[-cix]>0.5)>1) and cond_ix[-cix][0]>ci_threshold:
			keep[pi[:,-cix]>0.5]=False
	return keep,cond_ix

def find_singulars(panel,X):
	"""Returns a list with True for variables that cause singularity and False otherwise.
	for the main regression, singularity is detected by the constraints module"""
	N,T,k=X.shape
	XXCorrel=correl(X,panel)
	keep=np.all(np.isnan(XXCorrel)==False,0)
	keep=keep*np.all((np.abs(np.triu(XXCorrel,1))>0.99)==False,0)
	x_dev=deviation(panel, X)
	var_x=np.sum(np.sum(x_dev**2,0),0)
	keep=keep*(var_x>0)#remove constants
	keep[0]=True#allways keep the first constant term
	return keep,XXCorrel

def adf_test(panel,ll,p):
	"""Returns the augmented dickey fuller test statistic and critical value"""
	N,T,k=panel.X.shape
	y=ll.Y_st
	yl1=cf.roll(y,1,1)
	dy=y-yl1
	date_var=np.arange(T).reshape((T,1))*panel.included[3]	#date count
	X=np.concatenate((panel.included[3],date_var,yl1),2)
	dyL=[]
	for i in range(p):
		dyL.append(cf.roll(dy,i+1,1))
	dyL=np.concatenate(dyL,2)
	date_var=(date_var>p+1)
	X=np.concatenate((X,dyL),2)
	X=X*date_var
	dy=dy*date_var
	X[:,0:panel.lost_obs+10]=0
	keep,c_ix=singular_elim(panel,X)
	if not np.all(keep[0:3]):
		return 'NA','NA','NA'
	beta,se_robust,se=OLS(panel,X[:,:,keep],dy,robust_se_lags=10,c=date_var)
	adf_stat=beta[2]/se_robust[2]
	critval=adf_crit_values(panel.NT,True)
	res=np.append(adf_stat,critval)
	return res

def goodness_of_fit(ll,standarized):
	panel=ll.panel
	if standarized:
		s_res=panel.var(ll.e_RE)
		s_tot=panel.var(ll.Y_st)
	else:
		s_res=panel.var(ll.u)
		s_tot=panel.var(panel.Y)		
	r_unexpl=s_res/s_tot
	Rsq=1-r_unexpl
	Rsqadj=1-r_unexpl*(panel.NT-1)/(panel.NT-panel.args.n_args-1)
	panel.args.create_null_ll()
	LL_ratio_OLS=2*(ll.LL-panel.args.LL_OLS)
	LL_ratio=2*(ll.LL-panel.args.LL_null)
	return Rsq, Rsqadj, LL_ratio,LL_ratio_OLS


def breusch_godfrey_test(panel,ll, lags):
	"""returns the probability that err_vec are not auto correlated""" 
	e=ll.e_norm_centered
	X=ll.XIV_st
	N,T,k=X.shape
	X_u=X[:,lags:T]
	u=e[:,lags:T]
	c=panel.included[3][:,lags:T]
	for i in range(1,lags+1):
		X_u=np.append(X_u,e[:,lags-i:T-i],2)
	Beta,Rsq=OLS(panel,X_u,u,False,True,c=c)
	T=(panel.NT-k-1-lags)
	BGStat=T*Rsq
	rho=Beta[k:]
	ProbNoAC=1.0-chisq_dist(BGStat,lags)
	return ProbNoAC, rho, Rsq #The probability of no AC given H0 of AC.

def DurbinWatson(panel,ll):
	"""returns the probability that err_vec are not auto correlated""" 
	X=ll.XIV_st
	N,T,k=X.shape	
	e=ll.e_norm_centered
	c=panel.included[3][:,:-1]
	DW=np.sum((c*(e[:,1:]-e[:,:-1]))**2)/np.sum((e*panel.included[3])**2)
	return DW

def correlogram(panel,e,lags,center=False):
	N,T,k=e.shape	
	v=panel.var(e)
	if center:
		e=e-panel.mean(e)
	rho=np.zeros(lags+1)
	rho[0]=1
	for i in range(1,lags+1):
		a=panel.T_i-i-1>0
		incl=(a*panel.included[3])[:,i:,:]
		df=np.sum(incl)
		rho[i]=np.sum(incl*e[:,i:]*e[:,0:-i])/(v*df)
	return rho #The probability of no AC given H0 of AC.

def chisq_dist(X,df):
	"""Returns the probability of drawing a number
	less than X from a chi-square distribution with 
	df degrees of freedom"""
	retval=1.0-sc.gammaincc(df/2.0,X/2.0)
	return retval


def adf_crit_values(n,trend):
	"""Returns 1 and 5 percent critical values respectively. Interpolated from table in 
	Fuller, W. A. (1976). Introduction to Statistical Time Series. New York: John Wiley and Sons. ISBN 0-471-28715-6
	table is available at https://en.wikipedia.org/wiki/Augmented_Dickey%E2%80%93Fuller_test"""
	if trend:
		d={25:np.array([-3.75,-3.00]),50:np.array([-3.58,-2.93]),100:np.array([-3.51,-2.89]),
		   250:np.array([-3.46,-2.88]),500:np.array([-3.44,-2.87]),10000:np.array([-3.43,-2.86])}
	else:
		d={25:np.array([-4.38,-3.60]),50:np.array([-4.15,-3.50]),100:np.array([-4.04,-3.45]),
		   250:np.array([-3.99,-3.43]),500:np.array([-3.98,-3.42]),10000:np.array([-3.96,-3.41])}

	if n<25:
		print ("Warning: ADF critical values are not available for fewer than 25 observations")
		return (0,0)
	k=(25,50,100,250,500,10000)
	r=None
	for i in range(len(k)-1):
		if n>=k[i] and n<500:
			r=d[k[i]]+(n-k[i])*((d[k[i+1]]-d[k[i]])/(k[i+1]-k[i]))#interpolation
			return r
	if r is None:
		return d[10000]

def JB_normality_test(e,panel):
	"""Jarque-Bera test for normality. 
	returns the probability that a set of residuals are drawn from a normal distribution"""
	e=e[panel.included[3]]
	a=np.argsort(np.abs(e))[::-1]
	
	ec=e[a][int(0.001*len(e)):]
	
	df=len(ec)
	ec=ec-np.mean(ec)
	s=(np.sum(ec**2)/df)**0.5
	mu3=np.sum(ec**3)/df
	mu4=np.sum(ec**4)/df
	S=mu3/s**3
	C=mu4/s**4
	JB=df*((S**2)+0.25*(C-3)**2)/6.0
	p=1.0-chisq_dist(JB,2)
	return p


def correl(X,panel=None):
	"""Returns the correlation of X. Assumes three dimensional matrices. """
	if not panel is None:
		X=X*panel.included[3]
		N,T,k=X.shape
		N=panel.NT
		mean=np.sum(np.sum(X,0),0).reshape((1,k))/N
	else:
		N,k=X.shape
		mean=np.sum(X,0).reshape((1,k))/N
	cov=cf.dot(X,X)/N
	
	cov=cov-(mean.T*mean)
	stdx=(np.diag(cov)**0.5).reshape((1,k))
	stdx=(stdx.T*stdx)
	stdx[np.isnan(stdx)]=0
	corr=(stdx>0)*cov/(stdx+(stdx==0)*1e-100)
	corr[stdx<=0]=0
	
	return corr

def deviation(panel,X):
	N,T,k=X.shape
	x=X*panel.included[3]
	mean=np.sum(np.sum(x,0),0).reshape((1,1,k))/panel.NT
	return (X-mean)*panel.included[3]

def correl_2dim(X,Y=None,covar=False):
	"""Returns the correlation of X and Y. Assumes two dimensional matrixes. If Y is not supplied, the 
	correlation matrix of X is returned"""	
	if type(X)==list:
		X=Concat(X)
	single=Y is None
	if single:
		Y=X
	T,k=X.shape
	T,m=Y.shape
	X_dev=X-np.mean(X,0)
	Y_dev=Y-np.mean(Y,0)
	cov=np.dot(X_dev.T,Y_dev)
	if covar:
		return cov/(len(X)-1)
	stdx=np.sum(X_dev**2,0).reshape((1,k))**0.5
	if single:
		stdy=stdx
	else:
		stdy=np.sum(Y_dev**2,0).reshape((1,k))**0.5
	std_matr=stdx.T*stdy
	std_matr=std_matr+(std_matr==0)*1e-200
	corr=cov/std_matr
	if corr.shape==(1,1): 
		corr=corr[0][0]
	return corr

def get_singular_list(panel,XX):
	a,b=singular_elim(panel,XX)
	names=np.array(panel.input.X_names)[a==False]
	idx=np.array(range(len(a)))[a==False]
	s=', '.join([f"{names[i]}" for i in range(len(idx))])	
	return s

def OLS(panel,X,Y,add_const=False,return_rsq=False,return_e=False,c=None,robust_se_lags=0):
	"""runs OLS after adding const as the last variable"""
	if c is None:
		c=panel.included[3]
	N,T,k=X.shape
	NT=panel.NT
	if add_const:
		X=np.concatenate((c,X),2)
		k=k+1
	X=X*c
	Y=Y*c
	XX=cf.dot(X,X)
	XY=cf.dot(X,Y)
	try:
		beta=np.linalg.solve(XX,XY)
	except np.linalg.LinAlgError:
		s=get_singular_list(panel,X)
		raise RuntimeError("The following variables caused singularity runtime and must be removed: "+s)
	if return_rsq or return_e or robust_se_lags:
		e=(Y-cf.dot(X,beta))*c
		if return_rsq:
			v0=panel.var(e,included=c)
			v1=panel.var(Y,included=c)
			Rsq=1-v0/v1
			#Rsqadj=1-(v0/v1)*(NT-1)/(NT-k-1)
			return beta,Rsq
		elif return_e:
			return beta,e*c
		elif robust_se_lags:
			XXInv=np.linalg.inv(XX)
			se_robust,se,V=robust_se(panel,robust_se_lags,XXInv,X*e)
			return beta,se_robust.reshape(k,1),se.reshape(k,1)
	return beta

def OLS_simple(Y,X,addconst=False,residuals=True):
	"""Returns the OLS residuals if residuals. For use with two dimiensional arrays only"""
	if addconst:
		n=len(X)
		X=np.concatenate((np.ones((n,1)),X),1)
	XY=np.dot(X.T,Y)
	XX=np.dot(X.T,X)
	XXInv=np.linalg.inv(XX)
	beta=np.dot(XXInv,XY)
	e=Y-np.dot(X,beta)
	if residuals:
		return beta,e
	else:
		return beta

def newey_west_wghts(L,XErr):
	"""Calculates the Newey-West autocorrelation consistent weighting matrix. Either err_vec or XErr is required"""
	N,T,k=XErr.shape
	S=np.zeros((k,k))
	try:
		a=min(L,T)
	except:
		a=0
	for i in range(1,min(L,T)):
		w=1-(i+1)/(L)
		XX=cf.dot(XErr[:,i:],XErr[:,0:T-i])
		S+=w*(XX+XX.T)
	return S

def robust_cluster_weights(panel,XErr,cluster_dim,whites):
	"""Calculates the Newey-West autocorrelation consistent weighting matrix. Either err_vec or XErr is required"""
	N,T,k=XErr.shape
	if cluster_dim==0:#group cluster
		if N<=1:
			return 0
		mean=panel.mean(XErr,0)
	elif cluster_dim==1:#time cluster
		mean=random_effects.mean_time(panel,XErr,True)
		T,m,k=mean.shape
		mean=mean.reshape((T,k))
	S=cf.dot(mean,mean)-whites
	return S


def robust_se(panel,L,hessin,XErr,nw_only=True):
	"""Returns the maximum robust standard errors considering all combinations of sums of different combinations
	of clusters and newy-west"""
	w,W=sandwich_var(hessin,cf.dot(XErr,XErr))#whites
	nw,NW=sandwich_var(hessin,newey_west_wghts(L,XErr))#newy-west
	if panel.N>1:
		c0,C0=sandwich_var(hessin,robust_cluster_weights(panel,XErr, 0, w))#cluster dim 1
		c1,C1=sandwich_var(hessin,robust_cluster_weights(panel,XErr, 1, w))#cluster dim 2
	else:
		c0,c1,C0,C1=0,0,0*W,0*W
	v=np.array([
		nw,
		nw+c0,
		nw+c1,
		nw+c1+c0,
		w*0
	])
	V=np.array([
		NW,
		NW+C0,
		NW+C1,
		NW+C1+C0,
		W*0
	])
	V=V+W
	s=np.max(w+v,0)
	se_robust=np.maximum(s,0)**0.5
	i=np.argmax(np.sum(w+v,1))
	se_std=np.maximum(w,0)**0.5
	return se_robust,se_std,V[i]
	
def sandwich_var(hessin,V):
	hessinV=cf.dot(hessin,V)
	V=cf.dot(hessinV,hessin)
	v=np.diag(V)
	return v,V
	







#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import os
import multi_core
import traceback
import numpy as np
import time

def main(f):
	t=multi_core.transact(sys.stdin, sys.stdout)
	t.send(os.getpid())
	msg,(modules,s_id,f_node_name)=t.receive()
	f_node=open(f_node_name,'w')
	aliases=[]
	d_init=dict()
	for module,alias in modules:
		if alias=='':
			exec('import '+module,globals(),d_init)
			aliases.append(module,globals(),d_init)
		else:
			exec('import '+module +' as ' + alias,globals(),d_init)

	holdbacks=[]
	while 1:
		(msg,obj) = t.receive()
		response=None
		if msg==True:
			sys.exit()
			response=True
		elif msg=='static dictionary':#an initial dictionary to be used in the batch will be passed
			d=obj
			add_to_dict(d_init,d)
			response=True
		elif msg=='dynamic dictionary':#a dictionary to be used in the batch will be passed
			d=obj
			add_to_dict(d,d_init)
			d_list=list(d.keys())
			response=True
		elif msg=='expression evaluation':	
			sys.stdout = f_node
			exec(obj,globals(),d)
			sys.stdout = sys.__stdout__
			response=release_dict(d,d_list,holdbacks)
		elif msg=='holdbacks':
			holdbacks=obj  
			
		t.send(response)
		
def add_to_dict(to_dict,from_dict):
	for i in from_dict:
		to_dict[i]=from_dict[i]
		
def write(f,txt):
	f.write(str(txt)+'\n')
	f.flush()
	
def release_dict(d,d_list,holdbacks):
	"""Ensures that only new variables are returned"""
	response=dict()
	for i in d:
		if (not i in d_list) and (not i in holdbacks):
			response[i]=d[i]	
	return response

try: 
	f=open('slave_errors.txt','w')
	main(f)
except Exception as e:
	traceback.print_exc(file=f)
	f.flush()
	f.close()
	sys.exit()


import sys
sys.path.append('..//build//lib.win-amd64-3.5')
sys.path.append('..//build//lib.linux-x86_64-3.5')
import numpy
import ctypes as ct


gamma=numpy.array([0.5,0.2,-0.3])
lambda_=numpy.array([0.5,0.2,-0.3])



import cfunctions as c
bbb=c.bandinverse(gamma,lambda_,1500)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

version='1.1.13'

# Always prefer setuptools over distutils
from setuptools import setup, find_packages, Extension
# To use a consistent encoding
from codecs import open
from os import path
import subprocess
import sys
import time

try:
	import numpy as np
except:
	try:
		subprocess.check_call([sys.executable, "-m", "pip", "install", 'numpy'])
		time.sleep(2)
		import numpy as np
	except:
		print("""Warning: the python package numpy is not preinstalled. 
Without it the 'cfunctions' library will not be installed, which will affect performance.
Try uninstall paneltime, install numpy and install paneltime again""")

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.txt'), encoding='utf-8') as f:
	long_description = f.read()
try:
	ext=[]
	cfunc=Extension('cfunctions',sources=['paneltime/cfunctions.cpp'],include_dirs=[np.get_include()])
	ext=[cfunc]
except:
	print("""Warning: 'cfunctions' library not compiled. This may affect performance. 
If you are installing on a Windows machine, precompiled versions exist for python 3.5,3.6 and 3.7, so you should be fine""")
setup(
    name='paneltime',
    ext_modules=ext,
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,

    description='An efficient integrated panel and GARCH estimator',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/espensirnes/paneltime',

    # Author details
    author='Espen Sirnes',
    author_email='espen.sirnes@uit.no',

    # Choose your license
    license='GPL-3.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Researchers',
        'Topic :: Statistical Software :: time series estimation',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GPL-3.0 License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        ],

    # What does your project relate to?
    keywords='econometrics',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    
    
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    #**************************************************************************REMOVED>
    install_requires=['numpy >= 1.11','scipy','matplotlib','pymysql'],
    #**************************************************************************<REMOVED

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    #**************************************************************************REMOVED>
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    #     },
    #**************************************************************************<REMOVED 
    #
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    #**************************************************************************REMOVED>
    
    package_data={
        '': ['gui/*', '*.ico','gui/img/*.png',
		'cfunctions.cp35-win_amd64.pyd',
		'cfunctions.cp36-win_amd64.pyd',
		'cfunctions.cpython-35m-x86_64-linux-gnu.so',
		'cfunctions.cp37-win_amd64.pyd'],
        },
    include_package_data=True,
    #**************************************************************************<REMOVED
    #
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #**************************************************************************REMOVED>
    #data_files=[('my_data', ['data/data_file'])],
    #**************************************************************************<REMOVED
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'paneltime=paneltime:main',
            ],
        },
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from PIL import ImageTk, Image

def setbutton(parent,text,command,side=None,anchor=None,fill=None,bg=None,relief=None):
	btn=tk.Button(parent, text=text, command=command,anchor=anchor,bg=bg,relief=relief,bd=0)
	return btn


def save(subplot,save_file):
	fgr,axs=subplot
	fgr.savefig(save_file)
	axs.clear()
	
	
def display(chart,name,i,subplot,action=None,path=None):
	fgr,axs=subplot
	if path is None:
		path=chart.path
	f=open(path,'wb+')
	fgr.savefig(f)
	plot_to_chart(f,chart)
	axs.clear()
	f.close()
	
	chart.name=name
	chart.i=i
	chart.path=path
	chart.bind("<Button-1>", action)	

def display_from_img(chart,f,name,i,action=None):
	plot_to_chart(f,chart)
	chart.name=name
	chart.i=i
	chart.bind("<Button-1>", action)	
	

def plot_to_chart(f,chart):
	if hasattr(chart,'graph_file'):
		chart.graph_file.close()
	chart.graph_file=Image.open(f)
	img = ImageTk.PhotoImage(chart.graph_file,master=chart)
	chart.configure(image=img)
	chart.graph_img=img	
	
def fix_fname(s,i=None):
	if i is None:
		i=''
	else:
		i=str(i)
	if '.' in s:
		l=len(s.split('.')[-1])
		s=s[:-l]+i+s[-l:]
	else:
		s=s+i+'.jpg'
	return s





#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import functions as fu

monthdays=np.array([31,28,31,30,31,30,31,31,30,31,30,31]).reshape((1,12))
month_rng=np.arange(1,13).reshape((1,12))
monthdays=np.cumsum(monthdays,1)

def days_since(years=0,months=0,days=0,hours=0, minutes=0,seconds=0,base_year=2000):
	if (np.max(months)>12 or np.max(days)>31 
	    or np.min(days)<0 or  np.min(months)<0):
		raise RuntimeError("There is something wrong with a string variable that is formatted like a date-time variable")

	yr_rng=np.arange(min((np.min(years),base_year))-1,max((np.max(years),base_year)))
	n=len(days)
	k=len(yr_rng)
	yr_rng=yr_rng.reshape((1,k))
	
	#Calculating leap years
	leap_year=(yr_rng/4.0).astype(np.int)==yr_rng/4.0
	leap_year=leap_year*((yr_rng/100.0).astype(np.int)!=yr_rng/100.0)
	leap_year=(leap_year+((yr_rng/400.0).astype(np.int)==yr_rng/400.0))>0
	
	#Calculating number of days
	yr_sumdays=np.cumsum(leap_year+365,1)
	yr_days=np.sum((years-1==yr_rng)*yr_sumdays,1).reshape((n,1))
	month_days=np.sum((months-1==month_rng)*monthdays,1).reshape((n,1))
	month_days=month_days+(months>2)*np.sum((years==yr_rng)*leap_year,1).reshape((n,1))
	
	#subtracting january 1. of base year
	i=np.nonzero(yr_rng==base_year-1)[1][0]
	days_base_year=yr_sumdays[0,i]+1
	
	days_count=yr_days+month_days+days-days_base_year 
	days_count=days_count+time_fraction(hours, minutes, seconds)
	return days_count
	
	
def time_fraction(hours,minutes,seconds):
	if (np.max(hours)>24.0 or np.max(minutes)>60.0 or 
	    np.max(seconds)>60.0 or  np.min(hours)<0 
	    or np.min(minutes)<0 or  np.min(seconds)<0):
		raise RuntimeError("There is something wrong with a string variable that is formatted like a date-time variable")		
	return (hours/24.0)+(minutes/1440.0)+(seconds/86400.0)

	
def test():
	dates=np.loadtxt('../Input/date.csv',delimiter='.')
	days,months,years=dates[:,0:1],dates[:,1:2],dates[:,2:3]
	d=days_since_2000(years, months, days)
	fu.savevar(d,extension='csv')
	pass
	
	
#test()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from tkinter import filedialog
import os
import traceback
from multiprocessing import pool
from gui import gui_sql

class buttons:
	def __init__(self,win):
		self.win=win
		dirname=os.path.dirname(__file__)
		self.isevaluating=False
		imgs=[('run.png',self.run),
			  ('stop.png',self.stop),
			  ('statistics.png',self.statistics),
			  ('open_data.png',self.open),
			  ('open_data_sql.png',self.open_sql),
			  ('save.png',self.save),
			  ('save_as.png',self.save_as)]
		self.buttons={}
		for i in range(len(imgs)):
			name,action=imgs[i]
			short_name=name.replace('.png','')
			img = tk.PhotoImage(file =  os.path.join(dirname,'img',name),name=short_name,master=win.button_pane)
			b=tk.Button(win.button_pane, image = img,command=action, highlightthickness=0,bd=0,width=40)
			b.grid(row=0,column=i,sticky=tk.W)
			self.buttons[short_name]=[img,b,True]
		imgs=[['run_disabled.png',self.buttons['run'][1]],
			  ['stop_disabled.png',self.buttons['stop'][1]],
			  ['statistics_disabled.png',self.buttons['statistics'][1]]
			  ]
		for name,btn in imgs:#adding disabled-buttons
			img=tk.PhotoImage(file =  os.path.join(dirname,'img',name),name=name.replace('.png',''),master=win.button_pane)
			self.buttons[name.replace('.png','')]=[img,btn,False]
		self.gui_sql=None
	
	def run(self):
		self.win.data.save()
		if not self.buttons['run'][2] and False:
			return
		try:
			text=self.win.main_tabs.selected_tab_text()
			text.replace('start()','')
		except:
			return
		self.run_disable()
		self.pool = pool.ThreadPool(processes=1)
		self.process=self.pool.apply_async(self.win.exec, (text,),callback=self.run_enable)
		self.win.right_tabs.tabs.select(self.win.right_tabs.chart_tab)
		
	
	def stop(self):
		self.pool.close()
		
	def statistics(self):
		pass
	
	def open(self):
		p=self.win.data['current path']
		filename = filedialog.askopenfilename(initialdir=p,title="Open data file",
			filetypes = (("CSV", "*.csv")
	        ,("text", "*.txt")
			,("All files", "*.*") ))
		if not filename: 
			return
		p,f=os.path.split(filename)
		self.win.data['current path']=p
		exe_str=f"load('{filename}')"
		df=self.win.eval(exe_str)
		tree=self.win.right_tabs.data_tree
		tree.datasets.add(tree,f,df,filename,'df='+exe_str,self.win.main_tabs)

	

	
	def save(self):
		pass
	
	def save_as(self):
		pass
	
	def run_disable(self):
		self.new_image('run_disabled')
		self.new_image('stop')	
	
	def run_enable(self,events=None):
		self.new_image('run')
		self.new_image('stop_disabled')
		
	def new_image(self,img_name):
		img,buttn,enabled=self.buttons[img_name]
		self.buttons[img_name.replace('_disabled','')][2]=img_name[-9:]=='_disabled'
		buttn.configure(image=img)

		
		


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import tkinter as tk
from tkinter import ttk
from multiprocessing import pool

import numpy as np
import stat_functions as stat
from scipy import stats as scstats
from gui import gui_functions as guif
import os
import functions as fu



class process_charts(ttk.Frame):
	def __init__(self,window,master,main_tabs,tabs):
		style = ttk.Style()
		style.configure("TFrame", background='white')		
		ttk.Frame.__init__(self,master,style='new.TFrame')
		self.window=window
		self.ll=None	
		self.initialized=False
		self.subplot=tabs.subplot
		self.print_subplot=tabs.print_subplot
		self.img_tmp=tabs.img_tmp
		
	def get_images_for_storage(self):
		charts=[]
		for i in self.charts:
			charts.append((i.path,i.name))
		return charts
		
	def charts_from_stored(self,charts):
		self.add_content()
		if charts is None:
			return
		for i in range(len(charts)):
			path,name=charts[i]
			guif.display_from_img(self.charts[i],path,name,i)
		
	def add_content(self):
		self.n_charts=3
		self.columnconfigure(0,weight=1)
		for i in range(self.n_charts+1):
			self.rowconfigure(i,weight=1)
			

		tk.Label(self,text='Charts on normalized residuals:',bg='white',font='Tahoma 10 bold').grid(row=0,column=0)			

		self.charts=[]
		for i in range(self.n_charts):
			frm=tk.Frame(self,background='white')
			frm.rowconfigure(0,weight=1)
			frm.rowconfigure(1)
			frm.columnconfigure(0,weight=1)
			self.charts.append(tk.Label(frm,background='white'))
			self.charts[i].grid(row=0,column=0)	
			chart_path=os.path.join(os.getcwd(),'img',f'chart{i}.png')
			self.charts[i].path=fu.obtain_fname(chart_path)# self.img_tmp.TemporaryFile()
			guif.setbutton(frm, 'Save image', lambda: self.save(self.n_charts-i-1),bg='white').grid(row=1,column=0)
			frm.grid(row=i+1)
		
	def save(self,i):
		if not hasattr(self.charts[i],'graph_file') or not hasattr(self,'panel'):
			print('No graphics displayed yet')
			return
		name=self.charts[i].name
		f = tk.filedialog.asksaveasfile(mode='bw', defaultextension=".jpg",initialfile=f"{name}.jpg")		
		if f is None:
			return
		flst=[
			self.histogram,
			self.correlogram,
			self.correlogram_variance,
		]
		flst[i](self.ll,self.print_subplot,f)
		f.close()
		
	def initialize(self,panel):
		if not self.initialized:
			self.panel=panel
			self.add_content()
			self.initialized=True		
		
	def plot(self,ll):
		self.initialize(ll.panel)
		self.ll=ll
		self.histogram(ll,self.subplot)
		self.correlogram(ll,self.subplot)
		self.correlogram_variance(ll,self.subplot)	
		
		
	def histogram(self,ll,subplot,f=None):
		N,T,k=ll.panel.X.shape
		fgr,axs=subplot
		n=ll.e_norm_centered.shape[2]
		e=ll.e_norm_centered[self.panel.included[2]].flatten()
		N=e.shape[0]
		e=e.reshape((N,1))
		
		grid_range=4
		grid_step=0.05	
		h,grid=histogram(e,grid_range,grid_step)
		norm=scstats.norm.pdf(grid)*grid_step	
		
		axs.bar(grid,h,color='grey', width=0.025,label='histogram')
		axs.plot(grid,norm,'green',label='normal distribution')
		axs.legend(prop={'size': 6})
		name='Histogram - frequency'
		axs.set_title(name)
		if f is None:
			guif.display(self.charts[0],name,0,subplot)
		else:
			guif.save(subplot,f)

	def correlogram(self,ll,subplot,f=None):
		fgr,axs=subplot
		lags=20
		rho=stat.correlogram(self.panel, ll.e_norm_centered,lags)
		x=np.arange(lags+1)
		axs.bar(x,rho,color='grey', width=0.5,label='correlogram')
		name='Correlogram - residuals'
		axs.set_title(name)
		if f is None:
			guif.display(self.charts[1],name,1,subplot)
		else:
			guif.save(subplot,f)
		
	def correlogram_variance(self,ll,subplot,f=None):
		N,T,k=ll.panel.X.shape
		fgr,axs=subplot
		lags=20
		e2=ll.e_norm_centered**2
		e2=(e2-self.panel.mean(e2))*self.panel.included[3]
		rho=stat.correlogram(self.panel, e2,lags)
		x=np.arange(lags+1)
		axs.bar(x,rho,color='grey', width=0.5,label='correlogram')
		name='Correlogram - squared residuals'
		axs.set_title(name)
		if f is None:
			guif.display(self.charts[2],name,2,subplot)
		else:
			guif.save(subplot,f)
	
def histogram(x,grid_range,grid_step):
	N,k=x.shape
	grid_n=int(2*grid_range/grid_step)
	grid=np.array([i*grid_step-grid_range for i in range(grid_n)]).reshape((1,grid_n))
	ones=np.ones((N,1))
	x_u=np.concatenate((ones,x>=grid),1)
	x_l=np.concatenate((x<grid,ones),1)
	grid=np.concatenate((grid.flatten(),[grid[0,-1]+grid_step]))
	histogram=np.sum((x_u*x_l),0)
	if int(np.sum(histogram))!=N:
		raise RuntimeError('Error in histogram calculation')
	return histogram/N,grid

	



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#simulalte a panel data GARCH model

import numpy as np
import calculus_functions as cf
import functions as fu


class simulation:
	"""Creates an object that can simulate ARIMA-GARCH-timeseries data"""
	def __init__(self,N,T,beta,rho=[0.5],lmbda=[-0.5],psi=[0.5],gamma=[-0.5],omega=0.1,z=1,residual_sd=1,ID_sd=0,names=['x','const','Y','ID']):

		self.args,self.p,self.q,self.m,self.k,self.beta_len=self.new_args(beta,rho,lmbda,psi,gamma,omega,z)
		self.T=T
		self.names=['x','const','Y','ID']
		self.N=N
		self.residual_sd=residual_sd
		self.ID_sd=ID_sd
		self.max_lags=np.max((self.p,self.q,self.m,self.k))
		self.I=np.diag(np.ones(T))
		self.zero=self.I*0		
		self.AAR_1MA, self.GAR_1MA=matrices=self.set_garch_arch()
		
	def sim_many(self,n):
		for i in range(n):
			d=dict()
			X,Y,IDs=self.sim()
			save_dataset(X,Y,IDs,self.names,i)
			
	def de_sim(self,X,Y):
		args=self.args
		matrices=self.set_garch_arch_LL()
		AMA_1,AAR,AMA_1AR,GAR_1,GMA,GAR_1MA=matrices

		u=Y-cf.dot(X,args['beta'])
		e=cf.dot(AMA_1AR,u)
		e=u#fix this bug
		if self.m>0:
			h=np.log(e**2+args['z'])
			lnv_ARMA=cf.dot(GAR_1MA,h)
		else:
			lnv_ARMA=0
		lnv=args['omega']+lnv_ARMA# 'N x T x k' * 'k x 1' -> 'N x T x 1'

		v=np.exp(lnv)
		v_inv=np.exp(-lnv)

		LL=self.LL_const-0.5*np.sum((lnv+(e_REsq)*v_inv)*panel.included[3])	
		
		
	def sim_new(self):
		panel=self.panel
		X=panel.XIV
		matrices=set_garch_arch(panel,self.args.args_d)
		if matrices is None:
			return None		
		
		AMA_1,AMA_1AR,GAR_1,GAR_1MA=matrices
		(N,T,k)=X.shape
		#Idea for IV: calculate Z*u throughout. Mazimize total sum of LL. 
		u = panel.Y-cf.dot(X,self.args.args_d['beta'])
		e = cf.dot(AMA_1AR,u)
		e_RE = (e+self.re_obj_i.RE(e)+self.re_obj_t.RE(e))*panel.included[3]
		e_REsq = e_RE**2

		lnv_ARMA = self.garch(GAR_1MA, e)
		W_omega = cf.dot(panel.W_a, self.args.args_d['omega'])
		lnv = W_omega+lnv_ARMA# 'N x T x k' * 'k x 1' -> 'N x T x 1'
		#self.lnv0=lnv*1#debug
		grp = self.variance_RE(e_REsq)#experimental
		lnv+=grp
		self.dlnv_pos=(lnv<100)*(lnv>-100)
		lnv = np.maximum(np.minimum(lnv,100),-100)
		v = np.exp(lnv)*panel.a[3]
		v_inv = np.exp(-lnv)*panel.a[3]
		
		e_RE=e*v_inv**0.5
	
	
	def sim(self):
		args=self.args
		matrices=self.set_garch_arch_LL()
		AMA_1,AAR,AMA_1AR,GAR_1,GMA,GAR_1MA=matrices		
		if self.residual_sd==0:
			raise RuntimeError("Zero residual error not allowed.")
		e=np.random.normal(0,self.residual_sd,(self.N,self.T,1))
		if self.ID_sd>0:
			eRE=e+np.random.normal(0,self.ID_sd,(self.N,1,1))
		else:
			eRE=e
	
		self.eRE=cf.dot(self.AAR_1MA,eRE)#BUG u is currently not reused
	
	
		if self.m>0:
			h=np.log(self.eRE**2+args['z'])
			ID_eff=np.random.normal(0,1,(self.N,1,1))*args['mu']
			lnv=cf.dot(self.GAR_1MA,h)+ID_eff+args['omega']
		else:
			lnv=0	
	
		v=np.exp(lnv)
		v_inv=np.exp(-lnv)
	
		self.eRE_GARCH=self.eRE*v		

		X=np.random.normal(0,1,(self.N,self.T,self.beta_len-1))
		X=np.concatenate((np.ones((self.N,self.T,1)),X),2)
		Y_pred=cf.dot(X,args['beta'])
		Y=Y_pred+self.eRE_GARCH
		
		X=reshape(X,self.max_lags+1)
		Y=reshape(Y,self.max_lags+1)
		IDs=np.ones((self.N,self.T,1))*np.arange(self.N).reshape((self.N,1,1))
		IDs=reshape(IDs,self.max_lags+1)

		return X,Y,IDs

	def new_args(self,beta,rho,lmbda,psi,gamma,omega,mu,z):
		if rho is None:
			rho=[]
		if lmbda is None:
			lmbda=[]
		if psi is None:
			psi=[]
		if gamma is None:
			gamma=[]

		p=len(rho)
		q=len(lmbda)
		m=len(psi)
		k=len(gamma)
		beta_len=len(beta)		
		args=dict()
		args['beta']=np.array(beta).reshape((beta_len,1))
		args['omega']=omega
		args['rho']=np.array(rho)
		args['lambda']=np.array(lmbda)
		args['psi']=np.array(psi)
		args['gamma']=np.array(gamma)
		if m>0:
			args['mu']=mu
			args['z']=z
		else:
			args['mu']=0
			args['z']=0

		return args,p,q,m,k,beta_len

	
		
		
	def set_garch_arch(self):
		args,p,q,m,k=self.args,self.p,self.q,self.m,self.k
		AMA=cf.lag_matr(self.I,args['lambda'])
		X=-cf.lag_matr(-self.I,args['rho'])
		AAR_1=np.linalg.inv(X)
		AAR_1MA=cf.dot(AAR_1,AMA)
		X=-cf.lag_matr(self.I,args['gamma'])
		GAR_1=np.linalg.inv(X)
		GMA=cf.lag_matr(self.zero,args['psi'])	
		GAR_1MA=cf.dot(GAR_1,GMA)
		return AAR_1MA, GAR_1MA
	
	
	def set_garch_arch_LL(self):
		args,p,q,m,k=self.args,self.p,self.q,self.m,self.k
		X=cf.lag_matr(self.I,args['lambda'])
		try:
			AMA_1=np.linalg.inv(X)
		except:
			return None
		AAR=-cf.lag_matr(-self.I,args['rho'])
		AMA_1AR=cf.dot(AMA_1,AAR)
		X=-cf.lag_matr(-self.I,args['gamma'])
		try:
			GAR_1=np.linalg.inv(X)
		except:
			return None
		GMA=cf.lag_matr(self.zero,args['psi'])	
		GAR_1MA=cf.dot(GAR_1,GMA)
		return AMA_1,AAR,AMA_1AR,GAR_1,GMA,GAR_1MA	
	


def reshape(X,n):
	X=X[:,n:,:]
	return np.concatenate(X,0)
	



def save_dataset(X,Y,IDs,names,i):
	k=len(X[0])
	h=[]
	for j in range(len(X[0])):
		h.append(names[0]+str(j))
	h[0]=names[1]
	h=np.array([h])
	X=np.concatenate((h,X),0)
	Y=np.concatenate(([[names[2]]],Y),0)
	IDs=np.concatenate(([[names[3]]],IDs),0)
	data=np.concatenate((X,Y,IDs),1)
	fu.savevar(data,'/simulations/data_new'+str(i),'csv')	


#contains a numpy.ndarray derived class that manages zero observations in a panel

import numpy as np

class array(np.ndarray):
	def __new__(cls, array,included,ignore=False):
		obj = np.asarray(array).view(cls)
		cls.included=np.array(included)
		cls.ignore=ignore
		return obj

	
	def __add__(self,other):
		res=empty_handling(self,other)
		if not res is None:
			return res
		res=np.add(self,other)
		res=np.multiply(res,self.included[3])
		return res
		
	def __sub__(self,other):
		res=empty_handling(self,other)
		if not res is None:
			return res		
		res=np.subtract(self,other)
		res=np.multiply(res,self.included[3])
		return res		
		
	def __mul__(self,other):
		res=empty_handling(self,other)
		if not res is None:
			return res		
		res=np.multiply(self,other)
		res=np.multiply(res,self.included[3])
		return res		
		
	def __truediv__(self,other):
		res=empty_handling(self,other)
		if not res is None:
			return res		
		res=np.true_divide(self,other)
		res=np.multiply(res,self.included[3])
		return res
		
	def __eq__(self,other):
		res=empty_handling(self,other)
		if not res is None:
			return res		
		a=np.multiply(self,self.included[3])
		b=np.multiply(other,self.included[3])		
		res=np.equal(a,b)
		return res		


def empty_handling(self,other):
	if self.ignore:
		if len(other)==0:
			return self
		elif len(self)==0:
			return other
		else:
			return None
	else:
		if len(self)==0 or len(other)==0:
			return self[len(self):]	
		else:
			return None
		
a=array([2,4],[0,1])+array([8,10],[0,1])
f=a+np.array([])

a=0
